# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from planet_auth.oidc.api_clients.api_client import (
    OidcApiClient,
    EnricherFuncType,
    _RequestParamsType,
    _RequestAuthType,
)

# class RevocationAPIException(OidcApiClientException):
#
#   def __init__(self, message=None, raw_response=None):
#        super().__init__(message, raw_response)


class RevocationApiClient(OidcApiClient):
    """
    Low level API client for the OAuth Token Revocation endpoint. See
    RFC 7009 - OAuth 2.0 Token Revocation for protocol details.
    """

    def __init__(self, revocation_uri: str):
        """
        Create a new Revocation API Client.
        """
        super().__init__(endpoint_uri=revocation_uri)

    def _checked_revocation_call(self, params: _RequestParamsType, request_auth: Optional[_RequestAuthType]) -> None:
        self._checked_post(params, request_auth)
        # if response.content:
        #    # No payload expected on success. All HTTP and known json error
        #    # checks in base class.
        #    raise RevocationAPIException(
        #        message='Unexpected response from OIDC Revocation endpoint',
        #        raw_response=response)

    def _revoke_token(self, token: str, token_hint: str, auth_enricher: Optional[EnricherFuncType] = None) -> None:
        params = {
            "token": token,
            "token_type_hint": token_hint,
            # 'client_id': client_id # FIXME? Required? Part of enrichment?
        }
        request_auth = None
        if auth_enricher:
            params, request_auth = auth_enricher(params, self._endpoint_uri)
        self._checked_revocation_call(params, request_auth)

    def revoke_access_token(self, access_token: str, auth_enricher: Optional[EnricherFuncType] = None) -> None:
        """
        Revoke the specified access token.
        """
        self._revoke_token(access_token, "access_token", auth_enricher)

    def revoke_refresh_token(self, refresh_token: str, auth_enricher: Optional[EnricherFuncType] = None) -> None:
        """
        Revoke the specified refresh token.
        """
        return self._revoke_token(refresh_token, "refresh_token", auth_enricher)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from planet_auth.credential import Credential
from planet_auth.storage_utils import InvalidDataException, ObjectStorageProvider


class FileBackedOidcCredential(Credential):
    """
    Credential object for storing OAuth/OIDC tokens.
    """

    def __init__(self, data=None, credential_file=None, storage_provider: Optional[ObjectStorageProvider] = None):
        super().__init__(data=data, file_path=credential_file, storage_provider=storage_provider)

    def check_data(self, data):
        """
        Check that the supplied data represents a valid OAuth/OIDC token object.
        """
        super().check_data(data)
        if not data.get("access_token") and not data.get("id_token") and not data.get("refresh_token"):
            raise InvalidDataException(
                message="'access_token', 'id_token', or 'refresh_token' not found in file {}".format(self._file_path)
            )

    def access_token(self):
        """
        Get the current access token.
        """
        return self.lazy_get("access_token")

    def id_token(self):
        """
        Get the current ID token.
        """
        return self.lazy_get("id_token")

    def refresh_token(self):
        """
        Get the current refresh token.
        """
        return self.lazy_get("refresh_token")



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pyqrcode  # type: ignore
import time
from abc import ABC
from typing import List, Optional
from webbrowser import open_new

import planet_auth.logging.auth_logger

from planet_auth.auth_client import AuthClientException
from planet_auth.oidc.auth_client import (
    OidcAuthClientConfig,
    OidcAuthClient,
    OidcAuthClientWithNoneClientAuth,
)
from planet_auth.oidc.auth_client_with_client_pubkey import (
    OidcAuthClientWithClientPubkey,
    OidcAuthClientWithPubKeyClientConfig,
)
from planet_auth.oidc.auth_client_with_client_secret import (
    OidcAuthClientWithClientSecretClientConfig,
    OidcAuthClientWithClientSecret_HttpBasicAuthEnrichment,
)
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.oidc.auth_client_default_authenticators import (
    OidcAuthClientWithRefreshingOidcTokenRequestAuthenticator,
)


auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class DeviceCodeAuthClientException(AuthClientException):
    pass


class DeviceCodeClientConfig(OidcAuthClientConfig):
    """
    Configuration required for [planet_auth.DeviceCodeAuthClient][]
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_device_code",
            "auth_client_class": DeviceCodeAuthClient,
            "display_name": "Device Code",
            "description": "OAuth2 Device Code Flow for a non-confidential client.",
            "config_hints": super().meta().get("config_hints"),
        }


class DeviceCodeAuthClientBase(OidcAuthClientWithRefreshingOidcTokenRequestAuthenticator, OidcAuthClient, ABC):
    def __init__(self, client_config: DeviceCodeClientConfig):
        super().__init__(client_config)
        self._devicecode_client_config = client_config

    def device_login_initiate(
        self,
        requested_scopes: List[str] = None,
        requested_audiences: List[str] = None,
        extra: Optional[dict] = None,
        **kwargs,
    ) -> dict:
        """
        Obtain tokens from the OIDC auth server using the Device Code OAuth
        flow.  This method initiates the device login process with the auth
        service.  A dictionary that complies to [RFC 8628](https://www.rfc-editor.org/rfc/rfc8628)
        will be return upon success that can be used by the client application
        to prompt the user.  After prompting the user, `device_login_complete()`
        should be called to complete the login process.

        Args:
            requested_scopes: a list of strings specifying the scopes to
                request.
            requested_audiences: a list of strings specifying the audiences
                to request.
            extra: a dict extra data to pass to the authorization server.
         Returns:
            A RFC 8628 compliant dictionary that can be used by the called
            to prompt the user to complete device authentication.
            This dictionary should be passed to `device_login_complete()`
            to finish the login process.
        """
        final_requested_scopes, final_requested_audiences, final_extra = self._apply_config_fallback(
            requested_scopes=requested_scopes, requested_audiences=requested_audiences, extra=extra
        )

        return self.device_authorization_client().request_device_code(
            client_id=self._oidc_client_config.client_id(),
            requested_scopes=final_requested_scopes,
            requested_audiences=final_requested_audiences,
            auth_enricher=self._client_auth_enricher,
            extra=final_extra,
        )

    def device_login_complete(self, initiated_login_data: dict) -> FileBackedOidcCredential:
        """
        Obtain tokens from the OIDC auth server using the Device Code OAuth
        flow. This method completes the process initiated by a call to `device_login_initiate()`.

        Args:
            initiated_login_data: The dictionary returned from a successful call to
                `device_login_initiate()`
         Returns:
            A FileBackedOidcCredential object
        """
        # Wait one poll interval before querying.
        poll_interval = initiated_login_data.get("interval", 5)  # Default specified in RFC 8628
        time.sleep(poll_interval)
        token_json = self.token_client().poll_for_token_from_device_code(
            client_id=self._oidc_client_config.client_id(),
            device_code=initiated_login_data.get("device_code"),
            timeout=initiated_login_data.get("expires_in"),
            poll_interval=poll_interval,
            auth_enricher=self._client_auth_enricher,
        )
        return FileBackedOidcCredential(token_json)

    def _oidc_flow_login(
        self,
        allow_open_browser: Optional[bool],
        allow_tty_prompt: Optional[bool],
        requested_scopes: Optional[List[str]],
        requested_audiences: Optional[List[str]],
        extra: Optional[dict],
        display_qr_code: bool = False,
        **kwargs,
    ) -> FileBackedOidcCredential:
        """
        Obtain tokens from the OIDC auth server using the Device Code OAuth
        flow.

        Args:
            requested_scopes: a list of strings specifying the scopes to
                request.
            requested_audiences: a list of strings specifying the audiences
                to request.
            extra: a dict extra data to pass to the authorization server.
            allow_open_browser: specify whether login is permitted to open
                a browser window.
            allow_tty_prompt: specify whether login is permitted to request
                input from the terminal.
            display_qr_code: specify whether a QR code should be displayed on the tty
         Returns:
            A FileBackedOidcCredential object
        """
        self._warn_password_kwarg(**kwargs)
        self._warn_ignored_kwargs(["username", "password", "client_id", "client_secret"], **kwargs)

        if not allow_tty_prompt:
            # Without a TTY to confirm the authorization code with the user,
            # the client using this library really must use device_login_initiate and
            # device_login_complete and manage the user experience.
            raise DeviceCodeAuthClientException(message="Terminal input is required for device code login.")

        device_auth_response = self.device_login_initiate(
            requested_scopes=requested_scopes,
            requested_audiences=requested_audiences,
            extra=extra,
        )
        # "verification_uri_complete" is optional under the RFC.
        # "verification_uri" and "user_code" are not.
        verification_uri_complete = device_auth_response.get("verification_uri_complete")
        verification_uri = device_auth_response.get("verification_uri")
        user_code = device_auth_response.get("user_code")

        if allow_open_browser:
            if verification_uri_complete:
                print(f"Opening browser to login.\n" f"Confirm the authorization code when prompted: {user_code}\n")
                auth_logger.debug(msg=f'Opening browser with authorization URL: "{verification_uri_complete}"\n')
                open_new(verification_uri_complete)
            else:
                print(
                    f"Opening browser to login site {verification_uri}\n"
                    f"Enter authorization code when prompted: {user_code}\n"
                )
                auth_logger.debug(msg=f'Opening browser with authorization URL : "{verification_uri}"\n')
                open_new(verification_uri)  # type: ignore
        else:
            print("Please activate your client.")
            if verification_uri_complete:
                print(
                    f"Visit the activation site:\n"
                    f"\n\t{verification_uri_complete}\n"
                    f"\nand confirm the authorization code:\n"
                    f"\n\t{user_code}\n"
                )
                qr_uri = verification_uri_complete
            else:
                print(
                    f"Visit the activation site:\n"
                    f"\n\t{verification_uri}\n"
                    f"\nand enter the authorization code:\n"
                    f"\n\t{user_code}\n"
                )
                qr_uri = verification_uri

            if display_qr_code:
                qr_code = pyqrcode.create(content=qr_uri, error="L")
                print(f"Or, scan the QR code to complete login:\n\n{qr_code.terminal()}\n")

        return self.device_login_complete(device_auth_response)

    def can_login_unattended(self) -> bool:
        # Always requires user interaction
        return False


class DeviceCodeAuthClient(DeviceCodeAuthClientBase, OidcAuthClientWithNoneClientAuth):
    """
    AuthClient implementation that implements the OAuth device code grant
    to obtain user tokens.  This implementation is for public clients that cannot
    maintain client confidentiality.
    """

    # The base and mix-ins pretty much do it all.


class DeviceCodeWithClientSecretClientConfig(DeviceCodeClientConfig, OidcAuthClientWithClientSecretClientConfig):
    """
    Configuration required for [planet_auth.DeviceCodeWithClientSecretAuthClient][]
    """

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_device_code_secret",
            "auth_client_class": DeviceCodeWithClientSecretAuthClient,
            "display_name": "Device Code (Client Secret)",
            "description": "OAuth2 Devivce Code Flow for a confidential client using a shared client secret for authentication.",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class DeviceCodeWithClientSecretAuthClient(
    DeviceCodeAuthClientBase, OidcAuthClientWithClientSecret_HttpBasicAuthEnrichment
):
    """
    AuthClient implementation that implements the OAuth device code grant
    to obtain user tokens.  This implementation is for confidential
    clients that use a client secret to protect the client confidentiality.
    """

    # The base and mix-ins pretty much do it all.
    def __init__(self, client_config: DeviceCodeWithClientSecretClientConfig):
        super().__init__(client_config)


class DeviceCodeWithPubKeyClientConfig(DeviceCodeClientConfig, OidcAuthClientWithPubKeyClientConfig):
    """
    Configuration required for [planet_auth.DeviceCodeWithPubKeyAuthClient][]
    """

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_device_code_pubkey",
            "auth_client_class": DeviceCodeWithPubKeyAuthClient,
            "display_name": "Device Code (Public Key)",
            "description": "OAuth2 Device Code Flow for a confidential client using public key authentication.",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class DeviceCodeWithPubKeyAuthClient(DeviceCodeAuthClientBase, OidcAuthClientWithClientPubkey):
    """
    AuthClient implementation that implements the OAuth device code grant
    to obtain user tokens.  This implementation is for confidential
    clients that use a public/private keypair to protect the client confidentiality.
    """

    # The base and mix-ins pretty much do it all.
    def __init__(self, client_config: DeviceCodeWithPubKeyClientConfig):
        super().__init__(client_config)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from planet_auth.auth_client import AuthClientException
from planet_auth.none.noop_auth import NoOpCredential
from planet_auth.oidc.auth_clients.client_validator import (
    OidcClientValidatorAuthClientConfig,
    OidcClientValidatorAuthClient,
)
from planet_auth.request_authenticator import ForbiddenRequestAuthenticator


TEST_AUTH_SERVER_BASE = "https://auth.unittest.planet.com"


class ClientValidatorAuthClientConfigTest(unittest.TestCase):
    def test_default_client_id(self):
        under_test = OidcClientValidatorAuthClientConfig(auth_server=TEST_AUTH_SERVER_BASE)
        self.assertEqual(OidcClientValidatorAuthClientConfig.INTERNAL_CLIENT_ID, under_test.client_id())
        under_test = OidcClientValidatorAuthClientConfig(
            client_id="_utest_client_id_", auth_server=TEST_AUTH_SERVER_BASE
        )
        self.assertEqual("_utest_client_id_", under_test.client_id())


class ClientValidatorAuthClientTest(unittest.TestCase):
    # Most of the ClientValidatorAuthClientTest is actually base class functionality,
    # and unit tested with the base class.
    def setUp(self):
        self.under_test = OidcClientValidatorAuthClient(
            OidcClientValidatorAuthClientConfig(auth_server=TEST_AUTH_SERVER_BASE)
        )

    def test_login_fails(self):
        with self.assertRaises(AuthClientException):
            self.under_test.login()

    def test_refresh_fails(self):
        with self.assertRaises(AuthClientException):
            self.under_test.refresh(refresh_token="utest_dummy_refresh_token")

    def test_default_request_authenticator_type(self):
        test_result = self.under_test.default_request_authenticator(credential=NoOpCredential())
        self.assertIsInstance(test_result, ForbiddenRequestAuthenticator)



import logging
import pyqrcode # type: ignore
from planet_auth import Auth


def prompt_user(init_login_info):
    print("Please activate your device.")
    print(
        "Visit the activation site:\n"
        "\n\t{}\n"
        "\nand enter the activation code:\n"
        "\n\t{}\n".format(init_login_info.get("verification_uri"), init_login_info.get("user_code"))
    )
    if init_login_info.get("verification_uri_complete"):  # "verification_url_complete" is optional under the RFC.
        qr_code = pyqrcode.create(content=init_login_info.get("verification_uri_complete"), error="L")
        print(
            "Or, scan the QR code with your mobile device to visit {}:\n\n{}\n".format(
                init_login_info.get("verification_uri_complete"), qr_code.terminal()
            )
        )


def main():
    logging.basicConfig(level=logging.DEBUG)
    # In memory initialization.
    # Profiles can also be used with "device code" client types.
    auth_ctx = Auth.initialize_from_config_dict(
        client_config={
            "client_type": "oidc_device_code",
            "auth_server": "__auth_server__",
            "client_id": "__client_id__",
            "scopes": ["planet", "offline_access", "openid", "profile"],
        },
        token_file="/secure_device_storage/device_token.json",
    )

    login_init_info = auth_ctx.device_login_initiate()
    prompt_user(login_init_info)
    # credential will also be saved the file configured above, and
    # the request authenticator will be updated with the credential.
    credential = auth_ctx.device_login_complete(login_init_info)
    print(f"Credential saved to file {credential.path()}")


if __name__ == "__main__":
    main()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from planet_auth.planet_legacy.legacy_api_key import FileBackedPlanetLegacyApiKey
from planet_auth.storage_utils import FileBackedJsonObjectException
from tests.test_planet_auth.util import tdata_resource_file_path


class TestLegacyCredential(unittest.TestCase):
    def test_asserts_valid(self):
        under_test = FileBackedPlanetLegacyApiKey(
            api_key=None, api_key_file=tdata_resource_file_path("keys/planet_legacy_test_credential.json")
        )
        under_test.load()
        self.assertEqual("test_legacy_api_key", under_test.legacy_api_key())

        with self.assertRaises(FileBackedJsonObjectException):
            under_test.set_data(None)

        with self.assertRaises(FileBackedJsonObjectException):
            under_test.set_data({"test": "missing required fields"})

    def test_construct_with_literals(self):
        under_test = FileBackedPlanetLegacyApiKey(api_key="test_literal_apikey", jwt="test_literal_jwt")
        self.assertEqual("test_literal_apikey", under_test.legacy_api_key())
        self.assertEqual("test_literal_jwt", under_test.legacy_jwt())

    def test_getters_from_file(self):
        under_test = FileBackedPlanetLegacyApiKey(
            api_key=None, api_key_file=tdata_resource_file_path("keys/planet_legacy_test_credential.json")
        )
        under_test.load()
        self.assertEqual("test_legacy_api_key", under_test.legacy_api_key())
        self.assertEqual("test_legacy_literal_jwt", under_test.legacy_jwt())



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from requests import Session, Response
from requests.adapters import HTTPAdapter
from requests.auth import AuthBase
from typing import Callable, Dict, Optional, Tuple
from urllib3.util.retry import Retry

from planet_auth.auth_client import AuthClientException
from planet_auth.constants import X_PLANET_APP_HEADER, X_PLANET_APP
from planet_auth.util import parse_content_type

EnricherPayloadType = Dict
# EnricherAudType = str
EnricherReturnType = Tuple[Dict, Optional[AuthBase]]
EnricherFuncType = Callable[[EnricherPayloadType, str], EnricherReturnType]

_RequestAuthType = AuthBase
_RequestParamsType = Dict  # Requests allows a lot more, but constrain our use.
_RequestResponseType = Response


class OidcApiClient(ABC):
    """
    Base class that provides utility functions common to interactions with
    any of the OIDC endpoints.
    """

    # Most of the users of OidcApiClient will need some sort of "enriched" auth,
    # that takes client auth parameters from some facet of the client itself.
    # Generally, this will be some combination of the client ID and secret
    # and may be a header or payload adjustment.   But sometimes, we just
    # need to use an Authorization header.
    # TODO: dog-food - use our own RequestAuthenticator like we do for the
    #  static API key auth client
    class TokenBearerAuth(AuthBase):
        def __init__(self, token):
            self._token = token

        def __call__(self, r):
            r.headers["Authorization"] = "Bearer " + self._token
            return r

    def __init__(self, endpoint_uri: str):
        self._endpoint_uri = endpoint_uri

        retry_strategy = Retry(total=3, backoff_factor=1, status_forcelist=[429], allowed_methods=["POST", "GET"])
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self._session = Session()
        self._session.mount("https://", adapter)
        # self._session.mount("http://", adapter)

    def __check_http_error(self, response: _RequestResponseType) -> None:
        if not response.ok:
            raise OidcApiClientException(
                message="HTTP error from OIDC endpoint at {}: {}: {}".format(
                    self._endpoint_uri, response.status_code, response.reason
                ),
                raw_response=response,
            )

    def __check_oidc_payload_json_error(self, response: _RequestResponseType) -> None:
        if response.content:
            ct = parse_content_type(response.headers.get("content-type"))
            if not ct["content-type"] == "application/json":
                return

            json_response = response.json()

            # Irritatingly, I've seen multiple error payload schemas.
            # Most should adhere to RFC 6749.
            if json_response.get("error"):
                raise OidcApiClientException(
                    message="Error from OIDC endpoint at {}: {}: {}".format(
                        self._endpoint_uri, json_response.get("error"), json_response.get("error_description")
                    ),
                    error_code=json_response.get("error"),
                    raw_response=response,
                )

            if json_response.get("errorCode"):
                raise OidcApiClientException(
                    message="Error from OIDC endpoint at {}: {}: {}".format(
                        self._endpoint_uri, json_response.get("errorCode"), json_response.get("errorSummary")
                    ),
                    error_code=json_response.get("errorCode"),
                    raw_response=response,
                )

    @staticmethod
    def __checked_json_response(response: _RequestResponseType) -> Dict:
        json_response = None
        if response.content:
            ct = parse_content_type(response.headers.get("content-type"))
            if not ct["content-type"] == "application/json":
                raise OidcApiClientException(
                    message='Expected json content-type, but got "{}"'.format(response.headers.get("content-type")),
                    raw_response=response,
                )
            json_response = response.json()
        if not json_response:
            raise OidcApiClientException(
                messsage="Response was not understood. Expected JSON response payload, but none was found.",
                raw_response=response,
            )
        return json_response

    def __check_response(self, response: _RequestResponseType) -> None:
        # Check for the json error first so we throw a more specific parsed
        # error if we understand it, regardless of HTTP status code.
        self.__check_oidc_payload_json_error(response)
        self.__check_http_error(response)

    def _checked_get(
        self, params: Optional[_RequestParamsType], request_auth: Optional[_RequestAuthType]
    ) -> _RequestResponseType:
        response = self._session.get(
            self._endpoint_uri,
            params=params,
            auth=request_auth,
            headers={"Accept": "application/json", X_PLANET_APP_HEADER: X_PLANET_APP},
        )
        self.__check_response(response)
        return response

    def _checked_post(
        self, params: Optional[_RequestParamsType], request_auth: Optional[_RequestAuthType]
    ) -> _RequestResponseType:
        response = self._session.post(
            self._endpoint_uri,
            # Note: is the data/params crossing confusing? This was born out
            #       of some OIDC services accepting either in some cases,
            #       and others not doing so.
            data=params,
            auth=request_auth,
            headers={
                "Content-Type": "application/x-www-form-urlencoded",
                "Accept": "application/json",
                X_PLANET_APP_HEADER: X_PLANET_APP,
            },
        )
        self.__check_response(response)
        return response

    def _checked_post_json_response(
        self, params: Optional[_RequestParamsType], request_auth: Optional[_RequestAuthType]
    ) -> Dict:
        return self.__checked_json_response(self._checked_post(params, request_auth))

    def _checked_get_json_response(
        self, params: Optional[_RequestParamsType], request_auth: Optional[_RequestAuthType]
    ) -> Dict:
        return self.__checked_json_response(self._checked_get(params, request_auth))


class OidcApiClientException(AuthClientException):
    def __init__(self, error_code=None, raw_response=None, **kwargs):
        super().__init__(**kwargs)
        self.raw_response = raw_response
        self.error_code = error_code



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from requests.auth import AuthBase
from requests.models import Response
from typing import Tuple, Optional
from unittest import mock

from planet_auth.oidc.api_clients.introspect_api_client import (
    IntrospectionApiClient,
    IntrospectionApiException,
    IntrospectionRejectionTokenException,
)

TEST_API_ENDPOINT = "https://blackhole.unittest.planet.com/introspect"
TEST_ACCESS_TOKEN = "__test_access_token__"
TEST_ID_TOKEN = "__test_id_token__"
TEST_REFRESH_TOKEN = "__test_refresh_token__"
INTROSPECTION_RESPONSE_VALID = {
    "active": True,
    "scope": "offline_access profile openid",
    "username": "mock_test_user@planet.com",
    "exp": 1651255371,
    "iat": 1651251771,
    "sub": "mock_test_user@planet.com",
    "aud": "https://api.planet.com/mock_audience",
    "iss": "https://account.planet.com/mock_issuer",
    "jti": "mock_token_id",
    "token_type": "Bearer",
    "client_id": "mock_client_id",
    "uid": "mock_user_id",
    "api_key": "PLAK_MyKey",
    "user_id": 123456,
    "organization_id": 1,
    "role_level": 100,
}
INTROSPECTION_RESPONSE_FAILED = {"active": False}

INTROSPECTION_RESPONSE_INVALID1 = {
    ## Invalid - "active" field is not boolean. Even though this says active: true,
    ## the format is not valid and validation should fail.
    "active": "True",
    "scope": "offline_access profile openid",
    "username": "mock_test_user@planet.com",
    "exp": 1651255371,
    "iat": 1651251771,
    "sub": "mock_test_user@planet.com",
    "aud": "https://api.planet.com/mock_audience",
    "iss": "https://account.planet.com/mock_issuer",
    "jti": "mock_token_id",
    "token_type": "Bearer",
    "client_id": "mock_client_id",
    "uid": "mock_user_id",
    "api_key": "PLAK_MyKey",
    "user_id": 123456,
    "organization_id": 1,
    "role_level": 100,
}

INTROSPECTION_RESPONSE_INVALID2 = {
    ## Invalid - "active" field is not boolean
    "active": "False",
    "scope": "offline_access profile openid",
    "username": "mock_test_user@planet.com",
    "exp": 1651255371,
    "iat": 1651251771,
    "sub": "mock_test_user@planet.com",
    "aud": "https://api.planet.com/mock_audience",
    "iss": "https://account.planet.com/mock_issuer",
    "jti": "mock_token_id",
    "token_type": "Bearer",
    "client_id": "mock_client_id",
    "uid": "mock_user_id",
    "api_key": "PLAK_MyKey",
    "user_id": 123456,
    "organization_id": 1,
    "role_level": 100,
}

INTROSPECTION_RESPONSE_INVALID3 = {
    ## Invalid - "active" field is missing
    "scope": "offline_access profile openid",
    "username": "mock_test_user@planet.com",
    "exp": 1651255371,
    "iat": 1651251771,
    "sub": "mock_test_user@planet.com",
    "aud": "https://api.planet.com/mock_audience",
    "iss": "https://account.planet.com/mock_issuer",
    "jti": "mock_token_id",
    "token_type": "Bearer",
    "client_id": "mock_client_id",
    "uid": "mock_user_id",
    "api_key": "PLAK_MyKey",
    "user_id": 123456,
    "organization_id": 1,
    "role_level": 100,
}


def noop_auth_enricher(raw_payload: dict, audience: str) -> Tuple[dict, Optional[AuthBase]]:
    return raw_payload, None


def mocked_validate_ok(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(INTROSPECTION_RESPONSE_VALID))
    return response


def mocked_validate_fail(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(INTROSPECTION_RESPONSE_FAILED))
    return response


def mocked_validate_invalid1(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(INTROSPECTION_RESPONSE_INVALID1))
    return response


def mocked_validate_invalid2(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(INTROSPECTION_RESPONSE_INVALID2))
    return response


def mocked_validate_invalid3(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(INTROSPECTION_RESPONSE_INVALID3))
    return response


class IntrospectApiClientTest(unittest.TestCase):
    @mock.patch("requests.sessions.Session.post", side_effect=mocked_validate_ok)
    def test_validate_access_token_valid_with_enricher(self, mock_post):
        under_test = IntrospectionApiClient(introspect_uri=TEST_API_ENDPOINT)
        validation_data = under_test.validate_access_token(TEST_ACCESS_TOKEN, noop_auth_enricher)
        self.assertEqual(INTROSPECTION_RESPONSE_VALID, validation_data)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_validate_ok)
    def test_validate_access_token_valid(self, mock_post):
        under_test = IntrospectionApiClient(introspect_uri=TEST_API_ENDPOINT)
        validation_data = under_test.validate_access_token(TEST_ACCESS_TOKEN, None)
        self.assertEqual(INTROSPECTION_RESPONSE_VALID, validation_data)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_validate_fail)
    def test_validate_access_token_expired(self, mock_post):
        under_test = IntrospectionApiClient(introspect_uri=TEST_API_ENDPOINT)
        with self.assertRaises(IntrospectionRejectionTokenException):
            under_test.validate_access_token(TEST_ACCESS_TOKEN, None)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_validate_invalid1)
    def test_validate_access_token_invalid_response1(self, mock_post):
        under_test = IntrospectionApiClient(introspect_uri=TEST_API_ENDPOINT)
        with self.assertRaises(IntrospectionApiException):
            under_test.validate_access_token(TEST_ACCESS_TOKEN, None)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_validate_invalid2)
    def test_validate_access_token_invalid_response2(self, mock_post):
        under_test = IntrospectionApiClient(introspect_uri=TEST_API_ENDPOINT)
        with self.assertRaises(IntrospectionApiException):
            under_test.validate_access_token(TEST_ACCESS_TOKEN, None)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_validate_invalid3)
    def test_validate_access_token_invalid_response3(self, mock_post):
        under_test = IntrospectionApiClient(introspect_uri=TEST_API_ENDPOINT)
        with self.assertRaises(IntrospectionApiException):
            under_test.validate_access_token(TEST_ACCESS_TOKEN, None)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_validate_ok)
    def test_validate_id_token(self, mock_post):
        under_test = IntrospectionApiClient(introspect_uri=TEST_API_ENDPOINT)
        validation_data = under_test.validate_id_token(TEST_ID_TOKEN, None)
        self.assertEqual(INTROSPECTION_RESPONSE_VALID, validation_data)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_validate_ok)
    def test_validate_refresh_token(self, mock_post):
        under_test = IntrospectionApiClient(introspect_uri=TEST_API_ENDPOINT)
        validation_data = under_test.validate_refresh_token(TEST_REFRESH_TOKEN, None)
        self.assertEqual(INTROSPECTION_RESPONSE_VALID, validation_data)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from planet_auth.request_authenticator import CredentialRequestAuthenticator
from planet_auth.static_api_key.static_api_key import FileBackedApiKey


class FileBackedApiKeyRequestAuthenticator(CredentialRequestAuthenticator):
    """
    Load a bearer token from a file just in time for the request.
    Perform local checks on the validity of the token and throw
    if we think it will fail.
    """

    def __init__(self, api_key_credential: FileBackedApiKey, **kwargs):
        super().__init__(credential=api_key_credential, **kwargs)

    def pre_request_hook(self):
        self._credential.lazy_reload()
        self._token_body = self._credential.api_key()
        self._token_prefix = self._credential.bearer_token_prefix()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Planet Authentication and Authorization CLI Utilities.

These utilities are thin wrappers for functionality implemented
by the underlying auth libraries.

"""

from .commands.cli.main import (
    cmd_plauth_embedded,
    cmd_plauth_login,
    cmd_plauth_reset,
    cmd_plauth_version,
)
from .commands.cli.planet_legacy_auth_cmd import (
    cmd_pllegacy,
    cmd_pllegacy_login,
    cmd_pllegacy_print_api_key,
    cmd_pllegacy_print_access_token,
)
from .commands.cli.oauth_cmd import (
    cmd_oauth,
    cmd_oauth_login,
    cmd_oauth_refresh,
    cmd_oauth_validate_access_token_local,
    cmd_oauth_validate_access_token_remote,
    cmd_oauth_validate_id_token_local,
    cmd_oauth_validate_id_token_remote,
    cmd_oauth_validate_refresh_token_remote,
    cmd_oauth_revoke_access_token,
    cmd_oauth_revoke_refresh_token,
    cmd_oauth_userinfo,
    cmd_oauth_discovery,
    cmd_oauth_list_scopes,
    cmd_oauth_print_access_token,
)
from .commands.cli.profile_cmd import (
    cmd_profile,
    cmd_profile_list,
    cmd_profile_create,
    # cmd_profile_edit,
    cmd_profile_copy,
    cmd_profile_set,
    cmd_profile_show,
)
from .commands.cli.jwt_cmd import (
    cmd_jwt,
    cmd_jwt_decode,
    cmd_jwt_validate_oauth,
)
from .commands.cli.options import (
    opt_api_key,
    opt_audience,
    opt_client_id,
    opt_client_secret,
    opt_extra,
    opt_human_readable,
    opt_issuer,
    opt_loglevel,
    opt_long,
    opt_open_browser,
    opt_organization,
    opt_password,
    opt_profile,
    opt_project,
    opt_refresh,
    opt_scope,
    opt_show_qr_code,
    opt_sops,
    opt_token,
    opt_token_file,
    opt_username,
    opt_yes_no,
)
from .commands.cli.util import recast_exceptions_to_click
from planet_auth_utils.constants import EnvironmentVariables
from planet_auth_utils.plauth_factory import PlanetAuthFactory
from planet_auth_utils.builtins import Builtins
from planet_auth_utils.profile import Profile
from planet_auth_utils.plauth_user_config import PlanetAuthUserConfig

__all__ = [
    "cmd_plauth_embedded",
    "cmd_plauth_login",
    "cmd_plauth_reset",
    "cmd_plauth_version",
    "cmd_jwt",
    "cmd_jwt_decode",
    "cmd_jwt_validate_oauth",
    "cmd_oauth",
    "cmd_oauth_login",
    "cmd_oauth_refresh",
    "cmd_oauth_validate_access_token_local",
    "cmd_oauth_validate_access_token_remote",
    "cmd_oauth_validate_id_token_local",
    "cmd_oauth_validate_id_token_remote",
    "cmd_oauth_validate_refresh_token_remote",
    "cmd_oauth_revoke_access_token",
    "cmd_oauth_revoke_refresh_token",
    "cmd_oauth_userinfo",
    "cmd_oauth_discovery",
    "cmd_oauth_list_scopes",
    "cmd_oauth_print_access_token",
    "cmd_pllegacy",
    "cmd_pllegacy_login",
    "cmd_pllegacy_print_api_key",
    "cmd_pllegacy_print_access_token",
    "cmd_profile",
    "cmd_profile_list",
    "cmd_profile_create",
    # "cmd_profile_edit",
    "cmd_profile_copy",
    "cmd_profile_set",
    "cmd_profile_show",
    "opt_api_key",
    "opt_audience",
    "opt_client_id",
    "opt_client_secret",
    "opt_extra",
    "opt_human_readable",
    "opt_issuer",
    "opt_loglevel",
    "opt_long",
    "opt_open_browser",
    "opt_organization",
    "opt_password",
    "opt_profile",
    "opt_project",
    "opt_refresh",
    "opt_scope",
    "opt_show_qr_code",
    "opt_sops",
    "opt_token",
    "opt_token_file",
    "opt_username",
    "opt_yes_no",
    "recast_exceptions_to_click",
    #
    "Builtins",
    "EnvironmentVariables",
    "PlanetAuthFactory",
    "Profile",
    "PlanetAuthUserConfig",
]



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional
from planet_auth_utils.builtins_provider import BuiltinConfigurationProviderInterface


class MockStagingEnv:
    PRIMARY_OAUTH_AUTHORITY_USERS = {
        "_comment": "OIDC/OAuth server used by Public API endpoints",
        "auth_server": "https://mock-login1.staging.utest.planet.com/",
        "audiences": ["https://mock-api.staging.utest.planet.com/"],
    }
    PRIMARY_OAUTH_AUTHORITY_M2M = {
        "_comment": "OIDC/OAuth server used by Public API endpoints",
        "auth_server": "https://mock-login2.staging.utest.com/auth/realms/m2m",
        "audiences": ["https://mock-api.staging.utest.planet.com/"],
    }
    LEGACY_AUTH_AUTHORITY = {
        "_comment": "Planet legacy JWT auth server used by Public API endpoints",
        "legacy_auth_endpoint": "https://mock-api.staging.utest.planet.com/v0/auth/login",
    }
    TRUSTED_OAUTH_AUTHORITIES = [PRIMARY_OAUTH_AUTHORITY_USERS, PRIMARY_OAUTH_AUTHORITY_M2M]


class MockProductionEnv:
    PRIMARY_OAUTH_AUTHORITY_USERS = {
        "_comment": "OIDC/OAuth server used by Public API endpoints",
        "auth_server": "https://mock-login1.prod.utest.planet.com/",
        "audiences": ["https://mock-api.prod.utest.planet.com/"],
    }
    PRIMARY_OAUTH_AUTHORITY_M2M = {
        "_comment": "OIDC/OAuth server used by Public API endpoints",
        "auth_server": "https://mock-login2.prod.utest.com/auth/realms/m2m",
        "audiences": ["https://mock-api.prod.utest.planet.com/"],
    }
    LEGACY_AUTH_AUTHORITY = {
        "_comment": "Planet legacy JWT auth server used by Public API endpoints",
        "legacy_auth_endpoint": "https://mock-api.prod.utest.planet.com/v0/auth/login",
    }
    TRUESTED_OAUTH_AUTHORITIES = [PRIMARY_OAUTH_AUTHORITY_USERS, PRIMARY_OAUTH_AUTHORITY_M2M]


_BUILTIN_USER_CLIENT_PROD = {
    **MockProductionEnv.PRIMARY_OAUTH_AUTHORITY_USERS,
    "client_type": "oidc_device_code",
    # "client_type": "oidc_auth_code",
    "client_id": "TestMocke5yWHbTxzgMX9nFfn6jaEVy3",
    # "local_redirect_uri": "http://localhost:8080",
    "scopes": ["test_scope", "offline_access", "openid", "profile", "email"],
}

_BUILTIN_M2M_CLIENT_PROD = {
    **MockProductionEnv.PRIMARY_OAUTH_AUTHORITY_M2M,
    "client_type": "oidc_client_credentials_secret",
    "scopes": [],
    # "client_id": "__MUST_BE_USER_SUPPLIED__",
    # "client_secret": "__MUST_BE_USER_SUPPLIED__",
    # "scopes": ["test_scope"],
    # "audiences": [""]
}

_BUILTIN_USER_CLIENT_STAGING = {
    **MockStagingEnv.PRIMARY_OAUTH_AUTHORITY_USERS,
    "client_type": "oidc_device_code",
    # "client_type": "oidc_auth_code",
    "client_id": "TestMockYvR5jTcidWMNFUKV9AXdT6hI",
    # "local_redirect_uri": "http://localhost:8080", #
    "scopes": ["test_scope", "offline_access", "openid", "profile", "email"],
}

_BUILTIN_M2M_CLIENT_STAGING = {
    **MockStagingEnv.PRIMARY_OAUTH_AUTHORITY_M2M,
    "client_type": "oidc_client_credentials_secret",
    "scopes": [],
    # "client_id": "__MUST_BE_USER_SUPPLIED__",
    # "client_secret": "__MUST_BE_USER_SUPPLIED__",
    # "scopes": ["test_scope"],
}

_BUILTIN_LEGACY_AUTH_CLIENT_PROD = {
    **MockProductionEnv.LEGACY_AUTH_AUTHORITY,
    "client_type": "planet_legacy",
}

_BUILTIN_LEGACY_AUTH_CLIENT_STAGING = {
    **MockStagingEnv.LEGACY_AUTH_AUTHORITY,
    "client_type": "planet_legacy",
}

_NOOP_AUTH_CLIENT_CONFIG = {
    "client_type": "none",
}


class BuiltinConfigurationProviderMockTestImpl(BuiltinConfigurationProviderInterface):
    # fmt: off
    ##
    ## OAuth production environment profiles
    ##
    # Real
    BUILTIN_PROFILE_NAME_UTEST_USER          = "utest-user"
    BUILTIN_PROFILE_NAME_UTEST_M2M           = "utest-m2m"
    BUILTIN_PROFILE_NAME_UTEST_USER_STAGING  = "utest-user-staging"
    BUILTIN_PROFILE_NAME_UTEST_M2M_STAGING   = "utest-m2m-staging"
    # Aliases
    BUILTIN_PROFILE_ALIAS_UTEST_PROD          = "utest-prod"
    BUILTIN_PROFILE_ALIAS_UTEST_M2M_PROD      = "utest-m2m-prod"
    BUILTIN_PROFILE_ALIAS_UTEST_AUTH0_PROD    = "utest-user-prod"

    ##
    ## Profiles that use Planet's old (pre-OAuth) based auth protocol
    ##
    BUILTIN_PROFILE_NAME_UTEST_LEGACY         = "utest-legacy"
    BUILTIN_PROFILE_NAME_UTEST_LEGACY_STAGING = "utest-legacy-staging"

    ##
    ## Misc auth profiles
    ##
    BUILTIN_PROFILE_NAME_NONE    = "none"
    BUILTIN_PROFILE_ALIAS_UTEST_ALIAS_1 = "alias1"
    BUILTIN_PROFILE_ALIAS_UTEST_ALIAS_2 = "alais2"

    ##
    ## Default that should be used when no other selection has been made
    ##
    DEFAULT_PROFILE = BUILTIN_PROFILE_NAME_UTEST_USER

    _builtin_profile_auth_client_configs = {
        ## OAuth Client Configs
        BUILTIN_PROFILE_NAME_UTEST_USER          : _BUILTIN_USER_CLIENT_PROD,
        BUILTIN_PROFILE_NAME_UTEST_M2M           : _BUILTIN_M2M_CLIENT_PROD,
        BUILTIN_PROFILE_NAME_UTEST_USER_STAGING  : _BUILTIN_USER_CLIENT_STAGING,
        BUILTIN_PROFILE_NAME_UTEST_M2M_STAGING   : _BUILTIN_M2M_CLIENT_STAGING,

        # Planet Legacy Protocols
        BUILTIN_PROFILE_NAME_UTEST_LEGACY            : _BUILTIN_LEGACY_AUTH_CLIENT_PROD,
        BUILTIN_PROFILE_NAME_UTEST_LEGACY_STAGING    : _BUILTIN_LEGACY_AUTH_CLIENT_STAGING,

        # Misc
        BUILTIN_PROFILE_NAME_NONE              : _NOOP_AUTH_CLIENT_CONFIG,
    }

    _builtin_profile_aliases = {
        BUILTIN_PROFILE_ALIAS_UTEST_PROD                : BUILTIN_PROFILE_NAME_UTEST_USER,
        BUILTIN_PROFILE_ALIAS_UTEST_M2M_PROD            : BUILTIN_PROFILE_NAME_UTEST_M2M,
        BUILTIN_PROFILE_ALIAS_UTEST_AUTH0_PROD          : BUILTIN_PROFILE_NAME_UTEST_USER,
        BUILTIN_PROFILE_ALIAS_UTEST_ALIAS_1             : BUILTIN_PROFILE_NAME_UTEST_USER,
        BUILTIN_PROFILE_ALIAS_UTEST_ALIAS_2             : BUILTIN_PROFILE_ALIAS_UTEST_ALIAS_1,
    }

    _builtin_profile_default_by_client_type = {
        "oidc_device_code"               : BUILTIN_PROFILE_NAME_UTEST_USER,
        "oidc_auth_code"                 : BUILTIN_PROFILE_NAME_UTEST_USER,
        "oidc_client_credentials_secret" : BUILTIN_PROFILE_NAME_UTEST_M2M,
        "planet_legacy"                  : BUILTIN_PROFILE_NAME_UTEST_LEGACY,
    }

    _builtin_trust_realms: Dict[str, Optional[List[dict]]] = {
        "STAGING": MockStagingEnv.TRUSTED_OAUTH_AUTHORITIES,
        "PRODUCTION": MockProductionEnv.TRUESTED_OAUTH_AUTHORITIES,
        "CUSTOM": None,
    }
    # fmt: on

    def builtin_client_authclient_config_dicts(self) -> Dict[str, dict]:
        return self._builtin_profile_auth_client_configs

    def builtin_client_profile_aliases(self) -> Dict[str, str]:
        return self._builtin_profile_aliases

    def builtin_default_profile_by_client_type(self) -> Dict[str, str]:
        return self._builtin_profile_default_by_client_type

    def builtin_default_profile(self) -> str:
        return self.DEFAULT_PROFILE

    def builtin_trust_environment_names(self) -> List[str]:
        return list(BuiltinConfigurationProviderMockTestImpl._builtin_trust_realms.keys())

    def builtin_trust_environments(self) -> Dict[str, Optional[List[dict]]]:
        return BuiltinConfigurationProviderMockTestImpl._builtin_trust_realms



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import MagicMock

import freezegun
import pathlib
import tempfile
import unittest

from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.oidc.request_authenticator import (
    RefreshingOidcTokenRequestAuthenticator,
    RefreshOrReloginOidcTokenRequestAuthenticator,
)
from tests.test_planet_auth.unit.auth.util import StubOidcClientConfig, StubOidcAuthClient
from tests.test_planet_auth.util import tdata_resource_file_path

TEST_TOKEN_TTL = 8
TEST_TOKEN_AUDENCE = "__RefreshingOidcRequestAuthenticator_utest_audience__"
TEST_SIGNING_KEY_FILE = tdata_resource_file_path("keys/keypair1_priv_nopassword.test_pem")
TEST_SIGNING_PUBKEY_FILE = tdata_resource_file_path("keys/keypair1_pub_jwk.json")
TEST_AUTH_SERVER = "https://blackhole.unittest.planet.com/oauth2"

TEST_STUB_CLIENT_CONFIG = StubOidcClientConfig(
    auth_server=TEST_AUTH_SERVER,
    stub_authority_ttl=TEST_TOKEN_TTL,
    stub_authority_access_token_audience=TEST_TOKEN_AUDENCE,
    stub_authority_signing_key_file=TEST_SIGNING_KEY_FILE,
    stub_authority_pub_key_file=TEST_SIGNING_PUBKEY_FILE,
    scopes=["offline_access", "profile", "openid", "test_scope_1", "test_scope_2"],
)


class ORAUnitTestException(Exception):
    pass


class RefreshFailingStubOidcAuthClient(StubOidcAuthClient):
    def refresh(self, refresh_token, requested_scopes=None, extra=None, **kwargs):
        raise ORAUnitTestException("Forced test exception")


class RefreshingOidcRequestAuthenticatorTest(unittest.TestCase):
    def setUp(self):
        # we use the stub auth client for generating initial state, and the
        # mock auth client for probing authenticator behavior.
        self.stub_auth_client = StubOidcAuthClient(TEST_STUB_CLIENT_CONFIG)
        self.wrapped_stub_auth_client = MagicMock(wraps=self.stub_auth_client)
        self.refresh_failing_auth_client = RefreshFailingStubOidcAuthClient(TEST_STUB_CLIENT_CONFIG)
        self.wrapper_refresh_failing_auth_client = MagicMock(wraps=self.refresh_failing_auth_client)
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.tmp_dir_path = pathlib.Path(self.tmp_dir.name)

    def under_test_happy_path(self):
        credential_path = self.tmp_dir_path / "refreshing_oidc_authenticator_test_token__with_refresh.json"
        test_credential = self.mock_auth_login_and_command_initialize(
            credential_path=credential_path, auth_client=self.stub_auth_client
        )
        return RefreshingOidcTokenRequestAuthenticator(
            credential=test_credential, auth_client=self.wrapped_stub_auth_client
        )

    def under_test_happy_path_in_memory(self):
        test_credential = self.mock_auth_login_and_command_initialize_in_memory(
            credential_path=None, auth_client=self.stub_auth_client
        )
        return RefreshingOidcTokenRequestAuthenticator(
            credential=test_credential, auth_client=self.wrapped_stub_auth_client
        )

    def under_test_no_refresh_token(self):
        credential_path = self.tmp_dir_path / "refreshing_oidc_authenticator_test_token__without_refresh.json"
        test_credential = self.mock_auth_login_and_command_initialize(
            credential_path=credential_path, auth_client=self.stub_auth_client, get_refresh_token=False
        )
        return RefreshingOidcTokenRequestAuthenticator(
            credential=test_credential, auth_client=self.wrapped_stub_auth_client
        )

    def under_test_no_auth_client(self):
        credential_path = self.tmp_dir_path / "refreshing_oidc_authenticator_test_token__no_client_provided.json"
        test_credential = self.mock_auth_login_and_command_initialize(
            credential_path=credential_path, auth_client=self.stub_auth_client
        )
        return RefreshingOidcTokenRequestAuthenticator(credential=test_credential, auth_client=None)

    def under_test_no_access_token(self):
        credential_path = self.tmp_dir_path / "refreshing_oidc_authenticator_test_token__no_access_token.json"
        test_credential = self.mock_auth_login_and_command_initialize(
            credential_path=credential_path, auth_client=self.stub_auth_client, get_access_token=False
        )
        return RefreshingOidcTokenRequestAuthenticator(
            credential=test_credential, auth_client=self.wrapped_stub_auth_client
        )

    def under_test_refresh_fails(self):
        credential_path = self.tmp_dir_path / "refreshing_oidc_authenticator_test_token__refresh_fails.json"
        test_credential = self.mock_auth_login_and_command_initialize(
            credential_path=credential_path, auth_client=self.refresh_failing_auth_client
        )
        return RefreshingOidcTokenRequestAuthenticator(
            credential=test_credential, auth_client=self.wrapper_refresh_failing_auth_client
        )

    @staticmethod
    def mock_auth_login_and_command_initialize(
        credential_path, auth_client, get_access_token=True, get_id_token=True, get_refresh_token=True
    ):
        # pretend to bootstrap client auth configuration on disk, the way
        # it may be used by a user in an interactive shell:

        # bash$ planet auth login
        initial_credential = auth_client.login(
            get_access_token=get_access_token, get_refresh_token=get_refresh_token, get_id_token=get_id_token
        )
        initial_credential.set_path(credential_path)
        initial_credential.save()
        # <planet auth login process exits>

        # bash$ planet <some API command>
        #       # sets up credential object to be lazy loaded.
        test_credential = FileBackedOidcCredential(credential_file=credential_path)
        #       # The command would then use this credential and an
        #       # authenticator to interact with a planet API. Take it away,
        #       # test case...
        return test_credential

    @staticmethod
    def mock_auth_login_and_command_initialize_in_memory(
        credential_path, auth_client, get_access_token=True, get_id_token=True, get_refresh_token=True
    ):
        # pretend to bootstrap client auth in memory, the way
        # it may be used in a notebook or some other environment
        # where saving to disk is not available.
        # TODO: is this how you would go about this?

        # bash$ planet auth login would do this
        initial_credential = auth_client.login(
            get_access_token=get_access_token, get_refresh_token=get_refresh_token, get_id_token=get_id_token
        )
        initial_credential.set_path(credential_path)
        initial_credential.save()  # Should do nothing.

        # <planet auth login process would exit. We assume an in memory user does not.>
        return initial_credential

    @staticmethod
    def mock_api_call(under_test):
        # We don't need to actually mock making an HTTP API call. That is all
        # unit tested with the base class, and the OIDC authenticators don't
        # actually touch that. They only interact via the base class
        # pre_request_hook()
        under_test.pre_request_hook()

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_happy_path_1(self, frozen_time):
        # The first API call should trigger a token load from disk.
        # If the token is current, the auth client should be untouched.
        under_test = self.under_test_happy_path()

        self.assertIsNone(under_test._credential.data())
        self.mock_api_call(under_test)
        self.assertIsNotNone(under_test._credential.data())
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        access_token_t1 = under_test._credential.access_token()

        # inside the refresh window, more access should not refresh
        self.mock_api_call(under_test)
        self.mock_api_call(under_test)
        self.mock_api_call(under_test)
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        access_token_t2 = under_test._credential.access_token()
        self.assertEqual(access_token_t1, access_token_t2)

        # When the token reaches the 3/4 life, the authenticator should
        # attempt a token refresh
        frozen_time.tick(((3 * TEST_TOKEN_TTL) / 4) + 2)
        self.mock_api_call(under_test)
        self.assertEqual(1, under_test._auth_client.refresh.call_count)
        access_token_t3 = under_test._credential.access_token()
        self.assertNotEqual(access_token_t1, access_token_t3)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_happy_path_1_in_memory(self, frozen_time):
        # The first API call should trigger a token load from disk.
        # If the token is current, the auth client should be untouched.
        under_test = self.under_test_happy_path_in_memory()

        under_test._credential.check()  # Should have a fully valid credential
        self.mock_api_call(under_test)
        self.assertIsNotNone(under_test._credential.data())
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        access_token_t1 = under_test._credential.access_token()

        # inside the refresh window, more access should not refresh
        self.mock_api_call(under_test)
        self.mock_api_call(under_test)
        self.mock_api_call(under_test)
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        access_token_t2 = under_test._credential.access_token()
        self.assertEqual(access_token_t1, access_token_t2)

        # When the token reaches the 3/4 life, the authenticator should
        # attempt a token refresh
        frozen_time.tick(((3 * TEST_TOKEN_TTL) / 4) + 2)
        self.mock_api_call(under_test)
        self.assertEqual(1, under_test._auth_client.refresh.call_count)
        access_token_t3 = under_test._credential.access_token()
        self.assertNotEqual(access_token_t1, access_token_t3)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_happy_path_2(self, frozen_time):
        # The first API call should trigger a token load.
        # If the token is past the refresh time, a token refresh should be
        # attempted before the first use
        under_test = self.under_test_happy_path()

        frozen_time.tick(TEST_TOKEN_TTL + 2)
        self.mock_api_call(under_test)
        self.assertIsNotNone(under_test._credential.data())
        self.assertEqual(1, under_test._auth_client.refresh.call_count)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_refresh_fails(self, frozen_time):
        # Refresh could fail.  If it does, this should be buried and we
        # should try to carry on with the token we have.
        # (this is why we refresh before expiry, so transient failures
        # do not stop work). Of course, eventually the token will be no good,
        # and this is expected to generate API errors.
        under_test = self.under_test_refresh_fails()

        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        self.mock_api_call(under_test)
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        access_token_t1 = under_test._credential.access_token()

        frozen_time.tick(TEST_TOKEN_TTL + 2)

        self.mock_api_call(under_test)
        self.assertEqual(1, under_test._auth_client.refresh.call_count)
        access_token_t2 = under_test._credential.access_token()

        self.assertEqual(access_token_t1, access_token_t2)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_no_refresh_token(self, frozen_time):
        # if we have no refresh token, what happens when we expect to
        # refresh? It's not expected that a user use the refreshing
        # authenticator when not using a refresh token (there is another
        # authenticator for this), but you never know what people will do.
        # This should be treated like any other refresh failure (above).
        # We continue with what we have. (and while we would never expect a
        # call to refresh to work without a refresh token, that decision is
        # delegated to the auth server.)
        under_test = self.under_test_no_refresh_token()

        self.mock_api_call(under_test)
        access_token_t1 = under_test._credential.access_token()
        self.assertEqual(0, under_test._auth_client.refresh.call_count)

        frozen_time.tick(TEST_TOKEN_TTL + 2)

        self.mock_api_call(under_test)
        self.assertEqual(1, under_test._auth_client.refresh.call_count)
        access_token_t2 = under_test._credential.access_token()
        self.assertEqual(access_token_t1, access_token_t2)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_no_auth_client(self, frozen_time):
        # when no auth client is provided, just authenticate with what we
        # have.
        under_test = self.under_test_no_auth_client()

        self.mock_api_call(under_test)
        access_token_t1 = under_test._credential.access_token()

        frozen_time.tick(TEST_TOKEN_TTL + 2)

        self.mock_api_call(under_test)
        access_token_t2 = under_test._credential.access_token()
        self.assertEqual(access_token_t1, access_token_t2)

    def test_no_access_token(self):
        # Test credential has no access token, but it does have a refresh
        # token. We would not expect clients that want to use APIs to have
        # such credentials, but you never know what people will try!
        # It's up to the API endpoint to decide if "no auth" is valid.
        # For some endpoints, this may be valid (e.g. public discovery
        # endpoints.)
        #
        # This is not expected to be a normal path, but we should behave.
        # It allows an application to potentially be bootstrap with just
        # a refresh token.
        under_test = self.under_test_no_access_token()
        access_token_t1 = under_test._credential.access_token()
        self.assertIsNone(access_token_t1)
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        self.mock_api_call(under_test)
        self.assertEqual(1, under_test._auth_client.refresh.call_count)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_out_of_band_update_1(self, frozen_time):
        # Out of band credential update.  We expect the refresher to reload
        # the credential file without attempting a refresh.
        under_test = self.under_test_happy_path()

        self.assertIsNone(under_test._credential.data())
        self.mock_api_call(under_test)
        self.assertIsNotNone(under_test._credential.data())
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        access_token_t1 = under_test._credential.access_token()

        frozen_time.tick(TEST_TOKEN_TTL + 2)

        credential_t1 = under_test._credential
        oob_credential = self.stub_auth_client.refresh(credential_t1.refresh_token())
        oob_credential.set_path(credential_t1.path())
        oob_credential.save()
        access_token_oob = oob_credential.access_token()

        # The new token is within TTL, and a refresh should not occure
        self.mock_api_call(under_test)
        access_token_t2 = under_test._credential.access_token()
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        self.assertNotEqual(access_token_t1, access_token_t2)
        self.assertEqual(access_token_oob, access_token_t2)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_out_of_band_update_2(self, frozen_time):
        # Out of band credential update.  Something updated the credential
        # on disk underneath us. We expect the refresher to reload
        # the credential file without attempting a refresh.
        under_test = self.under_test_happy_path()

        self.assertIsNone(under_test._credential.data())
        self.mock_api_call(under_test)
        self.assertIsNotNone(under_test._credential.data())
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        access_token_t1 = under_test._credential.access_token()

        credential_t1 = under_test._credential
        oob_credential = self.stub_auth_client.refresh(credential_t1.refresh_token())
        oob_credential.set_path(credential_t1.path())
        oob_credential.save()
        access_token_oob = oob_credential.access_token()

        frozen_time.tick(TEST_TOKEN_TTL + 2)

        # The new token is outside TTL, even after the token is reloaded, this
        # should be detected and a refresh should still occur.
        self.mock_api_call(under_test)
        access_token_t2 = under_test._credential.access_token()
        self.assertEqual(1, under_test._auth_client.refresh.call_count)
        self.assertNotEqual(access_token_t1, access_token_t2)
        self.assertNotEqual(access_token_oob, access_token_t2)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_side_band_update_credential(self, frozen_time):
        # a side-band update of the credential should reset our "refresh at" time,
        # cascading behavior from there.  Where the above happens when someone
        # modifies a credential on disk, this happens when something in memory
        # provides us with a new credential, but it not the Authenticator's internal
        # self refreshing capabilities.
        under_test = self.under_test_happy_path()

        # The happy path under test primes brand new but *unloaded* credential that has been saved to disk.
        self.assertIsNone(under_test._credential.data())
        self.assertEqual(0, under_test._refresh_at)

        self.mock_api_call(under_test)  # Loads a valid token JIT.

        self.assertNotEqual(0, under_test._refresh_at)
        initial_access_token = under_test._credential.access_token()
        self.assertIsNotNone(initial_access_token)
        self.assertEqual(0, under_test._auth_client.refresh.call_count)

        # Get the sideband credential when we would have expected a refresh to probe behavior.
        frozen_time.tick(TEST_TOKEN_TTL + 2)
        sideband_credential_path = self.tmp_dir_path / "test_sideband.json"
        sideband_credential = self.mock_auth_login_and_command_initialize(
            credential_path=sideband_credential_path, auth_client=self.stub_auth_client
        )

        under_test.update_credential(sideband_credential)

        self.assertEqual(0, under_test._refresh_at)

        self.mock_api_call(under_test)  # Loads a valid token JIT.

        self.assertNotEqual(0, under_test._refresh_at)
        current_access_token = under_test._credential.access_token()
        self.assertEqual(sideband_credential.access_token(), current_access_token)
        self.assertNotEqual(initial_access_token, current_access_token)
        self.assertEqual(0, under_test._auth_client.refresh.call_count)

    # def test_side_band_update_credential_in_memory(self):
    #     under_test = self.under_test_happy_path_in_memory()

    def test_side_band_update_credential_data(self):
        # Similar to above, the credential is updated by a
        # side-band call to update the data.
        under_test = self.under_test_happy_path()
        self.mock_api_call(under_test)  # Should ensure that our old data is all primed and used
        initial_credential_data = under_test.credential().data()

        # abuse login to dummy up some other valid credential structure.
        sideband_credential = self.stub_auth_client.login(
            get_access_token=True, get_refresh_token=True, get_id_token=False
        )
        under_test.update_credential_data(sideband_credential.data())

        # update_credential_data() should leave is us in a freshly _load()'ed state.
        # It should not be necessary to simulate an API call for everything to be set
        current_credential_data = under_test._credential.data()
        self.assertNotEqual(current_credential_data, initial_credential_data)
        self.assertEqual(current_credential_data, sideband_credential.data())
        self.assertEqual(under_test._token_body, sideband_credential.access_token())
        self.assertNotEqual(0, under_test._refresh_at)
        self.assertNotEqual(0, under_test._credential._load_time)


class RefreshOrReloginOidcRequestAuthenticatorTest(unittest.TestCase):
    def setUp(self):
        # we use the stub auth client for generating initial state, and the
        # mock auth client for probing authenticator behavior.
        self.stub_auth_client = StubOidcAuthClient(TEST_STUB_CLIENT_CONFIG)
        self.mock_auth_client = MagicMock(wraps=self.stub_auth_client)
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.tmp_dir_path = pathlib.Path(self.tmp_dir.name)

    def under_test_with_refresh_token(self):
        credential_path = self.tmp_dir_path / "refreshing_or_relogin_oidc_authenticator_test_token__with_refresh.json"
        test_credential = self.mock_auth_login_and_command_initialize(credential_path)
        return RefreshOrReloginOidcTokenRequestAuthenticator(
            credential=test_credential, auth_client=self.mock_auth_client
        )

    def under_test_without_refresh_token(self):
        credential_path = (
            self.tmp_dir_path / "refreshing_or_relogin_oidc_authenticator_test_token__without_refresh.json"
        )
        test_credential = self.mock_auth_login_and_command_initialize(credential_path, get_refresh_token=False)
        return RefreshOrReloginOidcTokenRequestAuthenticator(
            credential=test_credential, auth_client=self.mock_auth_client
        )

    def under_test_no_auth_client(self):
        credential_path = (
            self.tmp_dir_path / "refreshing_or_relogin_oidc_authenticator_test_token__no_client_provided.json"
        )
        test_credential = self.mock_auth_login_and_command_initialize(credential_path)
        return RefreshOrReloginOidcTokenRequestAuthenticator(credential=test_credential, auth_client=None)

    def mock_auth_login_and_command_initialize(
        self, credential_path, get_access_token=True, get_id_token=True, get_refresh_token=True
    ):
        # pretend to bootstrap client auth configuration on disk, the way
        # it may be used by a user in an interactive shell:

        # bash$ planet auth login
        test_credential = self.stub_auth_client.login(
            get_access_token=get_access_token, get_refresh_token=get_refresh_token, get_id_token=get_id_token
        )
        test_credential.set_path(credential_path)
        test_credential.save()
        # <planet auth login process exits>

        # bash$ planet <some API command>
        #       # sets up credential object to be lazy loaded.
        test_credential = FileBackedOidcCredential(credential_file=credential_path)
        #       # The command would then use this credential and an
        #       # authenticator to interact with a planet API. Take it away,
        #       # test case...
        return test_credential

    def mock_api_call(self, under_test):
        # We don't need to actually mock making an HTTP API call. That is all
        # unit tested with the base class, and the OIDC authenticators don't
        # actually touch that. They only interact via the base class
        # pre_request_hook()
        under_test.pre_request_hook()

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_refresh_token_calls_refresh(self, frozen_time):
        under_test = self.under_test_with_refresh_token()

        self.assertIsNone(under_test._credential.data())
        self.mock_api_call(under_test)
        self.assertIsNotNone(under_test._credential.data())
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        self.assertEqual(0, under_test._auth_client.login.call_count)
        access_token_t1 = under_test._credential.access_token()

        frozen_time.tick(TEST_TOKEN_TTL + 2)

        self.mock_api_call(under_test)
        self.assertEqual(1, under_test._auth_client.refresh.call_count)
        self.assertEqual(0, under_test._auth_client.login.call_count)
        access_token_t2 = under_test._credential.access_token()
        self.assertNotEqual(access_token_t1, access_token_t2)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_no_refresh_token_calls_login(self, frozen_time):
        under_test = self.under_test_without_refresh_token()

        self.assertIsNone(under_test._credential.data())
        self.mock_api_call(under_test)
        self.assertIsNotNone(under_test._credential.data())
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        self.assertEqual(0, under_test._auth_client.login.call_count)
        access_token_t1 = under_test._credential.access_token()

        frozen_time.tick(TEST_TOKEN_TTL + 2)

        self.mock_api_call(under_test)
        self.assertEqual(0, under_test._auth_client.refresh.call_count)
        self.assertEqual(1, under_test._auth_client.login.call_count)
        access_token_t2 = under_test._credential.access_token()
        self.assertNotEqual(access_token_t1, access_token_t2)

    @freezegun.freeze_time(as_kwarg="frozen_time")
    def test_no_auth_client(self, frozen_time):
        # when no auth client is provided, just authenticate with what we
        # have.
        under_test = self.under_test_no_auth_client()

        self.mock_api_call(under_test)
        access_token_t1 = under_test._credential.access_token()

        frozen_time.tick(TEST_TOKEN_TTL + 2)

        self.mock_api_call(under_test)
        access_token_t2 = under_test._credential.access_token()
        self.assertEqual(access_token_t1, access_token_t2)



import logging
from planet_auth import Auth


def main():
    logging.basicConfig(level=logging.DEBUG)
    auth_ctx = Auth.initialize_from_profile(profile="my-custom-profile")
    credential = auth_ctx.login()  # Returned credential is also saved to disk in the profile directory.
    print("Credential saved to file {}".format(credential.path()))


if __name__ == "__main__":
    main()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
from requests.models import Response
from requests.auth import AuthBase
from typing import Tuple, Optional
from unittest import mock

from planet_auth.oidc.api_clients.device_authorization_api_client import (
    DeviceAuthorizationApiClient,
    DeviceAuthorizationApiException,
)

TEST_CLIENT_ID = "unittest_client_id"
TEST_API_ENDPOINT = "https://blackhole.unittest.planet.com/device_auth"

RESPONSE_VALID = {
    "device_code": "mock_device_code",
    "user_code": "mock_user_code",
    "verification_uri": "https://blackhole.unittest.planet.com/verify",
    "expires_in": 100,
}

RESPONSE_ERROR = {
    "errorCode": "invalid_client",
    "errorSummary": "Invalid value for 'client_id' parameter.",
    "errorLink": "invalid_client",
    "errorId": "oaxxxxzvfUZSabcdJmohb1234",
}

RESPONSE_INVALID_MISSING_DEVICE_CODE = RESPONSE_VALID.copy()
RESPONSE_INVALID_MISSING_DEVICE_CODE.pop("device_code")

RESPONSE_INVALID_MISSING_USER_CODE = RESPONSE_VALID.copy()
RESPONSE_INVALID_MISSING_USER_CODE.pop("user_code")

RESPONSE_INVALID_MISSING_VERIFICATION_URI = RESPONSE_VALID.copy()
RESPONSE_INVALID_MISSING_VERIFICATION_URI.pop("verification_uri")

RESPONSE_INVALID_MISSING_EXPIRES_IN = RESPONSE_VALID.copy()
RESPONSE_INVALID_MISSING_EXPIRES_IN.pop("expires_in")


def mocked_request_valid(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(RESPONSE_VALID))
    return response


def mocked_request_invalid_missing_device_code(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(RESPONSE_INVALID_MISSING_DEVICE_CODE))
    return response


def mocked_request_invalid_missing_user_code(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(RESPONSE_INVALID_MISSING_USER_CODE))
    return response


def mocked_request_invalid_missing_verification_uri(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(RESPONSE_INVALID_MISSING_VERIFICATION_URI))
    return response


def mocked_request_invalid_missing_expires_in(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(RESPONSE_INVALID_MISSING_EXPIRES_IN))
    return response


def noop_auth_enricher(raw_payload: dict, audience: str) -> Tuple[dict, Optional[AuthBase]]:
    return raw_payload, None


class DeviceAuthorizationApiClientTest(unittest.TestCase):
    @mock.patch("requests.sessions.Session.post", side_effect=mocked_request_valid)
    def test_request_device_code(self, mock_post):
        under_test = DeviceAuthorizationApiClient(device_authorization_uri=TEST_API_ENDPOINT)
        under_test.request_device_code(
            client_id=TEST_CLIENT_ID,
            requested_scopes=None,
            requested_audiences=None,
            auth_enricher=noop_auth_enricher,
            extra=None,
        )
        # No exception is a passed test

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_request_invalid_missing_device_code)
    def test_request_device_code_missing_device_code(self, mock_post):
        under_test = DeviceAuthorizationApiClient(device_authorization_uri=TEST_API_ENDPOINT)
        with self.assertRaises(DeviceAuthorizationApiException):
            under_test.request_device_code(
                client_id=TEST_CLIENT_ID,
                requested_scopes=["utest_scope_1"],
                requested_audiences=["utest_audience"],
                auth_enricher=None,
                extra=None,
            )

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_request_invalid_missing_user_code)
    def test_request_device_code_missing_user_code(self, most_post):
        under_test = DeviceAuthorizationApiClient(device_authorization_uri=TEST_API_ENDPOINT)
        with self.assertRaises(DeviceAuthorizationApiException):
            under_test.request_device_code(
                client_id=TEST_CLIENT_ID,
                requested_scopes=["utest_scope_1"],
                requested_audiences=["utest_audience"],
                auth_enricher=None,
                extra=None,
            )

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_request_invalid_missing_verification_uri)
    def test_request_device_code_missing_verification_uri(self, mock_post):
        under_test = DeviceAuthorizationApiClient(device_authorization_uri=TEST_API_ENDPOINT)
        with self.assertRaises(DeviceAuthorizationApiException):
            under_test.request_device_code(
                client_id=TEST_CLIENT_ID,
                requested_scopes=["utest_scope_1"],
                requested_audiences=["utest_audience"],
                auth_enricher=None,
                extra=None,
            )

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_request_invalid_missing_expires_in)
    def test_request_device_code_missing_expires_in(self, mock_post):
        under_test = DeviceAuthorizationApiClient(device_authorization_uri=TEST_API_ENDPOINT)
        with self.assertRaises(DeviceAuthorizationApiException):
            under_test.request_device_code(
                client_id=TEST_CLIENT_ID,
                requested_scopes=["utest_scope_1"],
                requested_audiences=["utest_audience"],
                auth_enricher=None,
                extra=None,
            )



import importlib.metadata


def on_config(config):
    """
    This is for injecting the package version into mkdocs
    config so it can be used in templates.
    """
    config["extra"]["site_version"] = importlib.metadata.version("planet-auth")
    return config



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import http.server

import httpx
import json
import pytest
import requests
import unittest

from planet_auth.auth_exception import AuthException
from planet_auth.constants import X_PLANET_APP_HEADER, X_PLANET_APP
from planet_auth.request_authenticator import ForbiddenRequestAuthenticator, SimpleInMemoryRequestAuthenticator
from planet_auth.static_api_key.request_authenticator import FileBackedApiKeyRequestAuthenticator
from planet_auth.static_api_key.static_api_key import FileBackedApiKey
from tests.test_planet_auth.util import background, find_free_port, is_cicd, tdata_resource_file_path

TEST_HEADER = "x-test-header"
TEST_PREFIX = "TEST"
TEST_TOKEN = "_test_bearer_token_"
TEST_TIMEOUT = 30


class _UnitTestRequestHandler(http.server.BaseHTTPRequestHandler):
    # def __init__(self, request, address, server):
    #     super().__init__(request, address, server)

    def do_GET(self):
        # Reflect the auth info back so unit test can check it.
        self.send_response(http.HTTPStatus.OK)
        self.send_header("Content-type", "application/json")
        self.end_headers()

        self.server.callback_raw_request_path = self.path
        response_payload = {
            "header_present": bool(self.headers.get(TEST_HEADER)),
            "header_value": self.headers.get(TEST_HEADER),
        }
        self.wfile.write(bytes(json.dumps(response_payload), "utf-8"))


@background
def handle_test_request_background(listen_port):
    # pylint: disable=W0108
    http_server = http.server.HTTPServer(
        ("localhost", listen_port), lambda request, address, server: _UnitTestRequestHandler(request, address, server)
    )
    http_server.timeout = TEST_TIMEOUT
    http_server.handle_request()


@pytest.mark.skipif(condition=is_cicd(), reason="Skipping tests that listen on a network port for CI/CD")
class RequestAuthenticatorNetworkTest(unittest.TestCase):
    def setUp(self):
        self.listen_port = find_free_port()
        handle_test_request_background(self.listen_port)
        self.test_server_url = "http://localhost:{}/test_request_lib".format(self.listen_port)

    def test_requests_auth(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=TEST_TOKEN, token_prefix=TEST_PREFIX, auth_header=TEST_HEADER
        )
        response = requests.get(self.test_server_url, auth=under_test, timeout=TEST_TIMEOUT)
        response_json = response.json()
        self.assertTrue(response_json.get("header_present"))
        self.assertEqual(TEST_PREFIX + " " + TEST_TOKEN, response_json.get("header_value"))

    def test_requests_no_auth(self):
        under_test = SimpleInMemoryRequestAuthenticator(token_body=None, token_prefix=None, auth_header=None)
        response = requests.get(self.test_server_url, auth=under_test, timeout=TEST_TIMEOUT)
        response_json = response.json()
        self.assertFalse(response_json.get("header_present"))
        self.assertIsNone(response_json.get("header_value"))

    def test_requests_no_prefix(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=TEST_TOKEN, token_prefix=None, auth_header=TEST_HEADER
        )
        response = requests.get(self.test_server_url, auth=under_test, timeout=TEST_TIMEOUT)
        response_json = response.json()
        self.assertTrue(response_json.get("header_present"))
        self.assertEqual(TEST_TOKEN, response_json.get("header_value"))

    def test_httpx_auth(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=TEST_TOKEN, token_prefix=TEST_PREFIX, auth_header=TEST_HEADER
        )
        response = httpx.get(self.test_server_url, auth=under_test)
        response_json = response.json()
        self.assertTrue(response_json.get("header_present"))
        self.assertEqual(TEST_PREFIX + " " + TEST_TOKEN, response_json.get("header_value"))

    def test_httpx_no_auth(self):
        under_test = SimpleInMemoryRequestAuthenticator(token_body=None, token_prefix=None, auth_header=None)
        response = httpx.get(self.test_server_url, auth=under_test)
        response_json = response.json()
        self.assertFalse(response_json.get("header_present"))
        self.assertIsNone(response_json.get("header_value"))

    def test_httpx_no_prefix(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=TEST_TOKEN, token_prefix=None, auth_header=TEST_HEADER
        )
        response = httpx.get(self.test_server_url, auth=under_test)
        response_json = response.json()
        self.assertTrue(response_json.get("header_present"))
        self.assertEqual(TEST_TOKEN, response_json.get("header_value"))

    # TODO: add aiohttp support


# The above is a better test, since it exercises how the base class interacts
# with the HTTP client libs requets and httpx and affects the resulting
# network traffic.  But, setting up a listener is problematic in some
# environments.  Testing the base class amounts to little more than testing
# setters. It doesn't validate that it interacts with the other libs to affect
# the network traffic the way we want it it do.  (We maintain this test anyway
# to hit our CI/CD coverage targets.)


def mock_httpx_get(under_test, headers=None):
    request = under_test.auth_flow(httpx._models.Request(url="http://unittest/", method="GET", headers=headers))
    return next(request)


def mock_requests_get(under_test, headers=None):
    request = requests.Request()
    if headers:
        request.headers.update(headers)
    under_test.__call__(request)
    return request


class RequestAuthenticatorNoNetworkTest(unittest.TestCase):
    def test_requests_auth(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=TEST_TOKEN, token_prefix=TEST_PREFIX, auth_header=TEST_HEADER
        )
        request = mock_requests_get(under_test)
        self.assertEqual(request.headers[TEST_HEADER], TEST_PREFIX + " " + TEST_TOKEN)

    def test_requests_no_auth(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=None, token_prefix=TEST_PREFIX, auth_header=TEST_HEADER
        )
        request = mock_requests_get(under_test)
        self.assertIsNone(request.headers.get(TEST_HEADER))

    def test_requests_no_prefix(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=TEST_TOKEN, token_prefix=None, auth_header=TEST_HEADER
        )
        request = mock_requests_get(under_test)
        self.assertEqual(request.headers[TEST_HEADER], TEST_TOKEN)

    def test_requests_x_planet_app_header_from_lib(self):
        under_test = SimpleInMemoryRequestAuthenticator()
        request = mock_requests_get(under_test)
        self.assertEqual(request.headers[X_PLANET_APP_HEADER], X_PLANET_APP)

    def test_requests_x_planet_app_header_from_app(self):
        under_test = SimpleInMemoryRequestAuthenticator()
        request = mock_requests_get(under_test, headers={X_PLANET_APP_HEADER: "unit-test-planet-client"})
        self.assertEqual(request.headers[X_PLANET_APP_HEADER], "unit-test-planet-client")

    def test_httpx_auth(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=TEST_TOKEN, token_prefix=TEST_PREFIX, auth_header=TEST_HEADER
        )
        request = mock_httpx_get(under_test)
        self.assertEqual(request.headers[TEST_HEADER], TEST_PREFIX + " " + TEST_TOKEN)

    def test_httpx_no_auth(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=None, token_prefix=TEST_PREFIX, auth_header=TEST_HEADER
        )
        request = mock_httpx_get(under_test)
        self.assertIsNone(request.headers.get(TEST_HEADER))

    def test_httpx_no_prefix(self):
        under_test = SimpleInMemoryRequestAuthenticator(
            token_body=TEST_TOKEN, token_prefix=None, auth_header=TEST_HEADER
        )
        request = mock_httpx_get(under_test)
        self.assertEqual(request.headers[TEST_HEADER], TEST_TOKEN)

    def test_httpx_x_planet_app_header_from_lib(self):
        under_test = SimpleInMemoryRequestAuthenticator()
        request = mock_httpx_get(under_test)
        self.assertEqual(request.headers[X_PLANET_APP_HEADER], X_PLANET_APP)

    def test_httpx_x_planet_app_header_from_app(self):
        under_test = SimpleInMemoryRequestAuthenticator()
        request = mock_httpx_get(under_test, headers={X_PLANET_APP_HEADER: "unit-test-planet-client"})
        self.assertEqual(request.headers[X_PLANET_APP_HEADER], "unit-test-planet-client")


class ForbiddenRequestAuthenticatorkTest(unittest.TestCase):
    # Since this request authenticator is supposed to pretty much prevent network
    # requests from being made, there is no need for separate Network/NoNetwork test cases.
    def setUp(self):
        self.under_test = ForbiddenRequestAuthenticator()

    def test_requests_auth(self):
        with self.assertRaises(AuthException):
            mock_requests_get(self.under_test)

    def test_httpx_auth(self):
        with self.assertRaises(AuthException):
            mock_httpx_get(self.under_test)


class CredentialRequestAuthenticatorTest(unittest.TestCase):
    def setUp(self) -> None:
        # This is the simplest concrete class that derives from CredentialRequestAuthenticatorTest
        self.test_credential_file_path_1 = tdata_resource_file_path("keys/static_api_key_test_credential.json")
        self.test_credential_file_path_2 = tdata_resource_file_path("keys/static_api_key_test_credential_2.json")
        self.test_credential_1 = FileBackedApiKey(api_key_file=self.test_credential_file_path_1)
        self.test_credential_2 = FileBackedApiKey(api_key_file=self.test_credential_file_path_2)
        self.under_test = FileBackedApiKeyRequestAuthenticator(api_key_credential=self.test_credential_1)

    def test_update_credential(self):
        self.assertIsNone(self.under_test._token_body)
        self.under_test.pre_request_hook()
        self.assertEqual("test_api_key", self.under_test._token_body)

        self.under_test.update_credential(self.test_credential_2)
        self.assertIsNone(self.under_test._token_body)
        self.under_test.pre_request_hook()
        self.assertEqual("test_api_key_2", self.under_test._token_body)



import logging
import httpx
from planet_auth import Auth


def main():
    logging.basicConfig(level=logging.DEBUG)
    auth_ctx = Auth.initialize_from_profile(profile="my-custom-profile")
    result = httpx.get(
        url="https://api.planet.com/basemaps/v1/mosaics",
        auth=auth_ctx.request_authenticator(),
        headers={"X-Planet-App": "httpx-example"},
    )
    print(result.status_code)
    print(result.json())


if __name__ == "__main__":
    main()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from planet_auth import AuthException


class MyException1(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg


class MyException1Sub1(MyException1):
    pass


class MyException1Sub2(MyException1):
    pass


class MyException2(Exception):
    def __init__(self, msg):
        super().__init__(msg)
        self.message = msg


class MyAuthException(AuthException):
    def __init__(self, message=None):
        super().__init__(message)


class MyAuthExceptionSub1(MyAuthException):
    pass


class MyAuthExceptionSub2(MyAuthException):
    pass


class MyAuthException2(AuthException):
    def __init__(self, message=None):
        super().__init__(message)


class MyAuthException3(AuthException):
    def __init__(self, message=None):
        super().__init__(message)


class TestAuthException(unittest.TestCase):
    def test_decorator_default(self):
        @AuthException.recast()
        def raise_my_exception1():
            raise MyException1(msg="test")

        with self.assertRaises(AuthException):
            raise_my_exception1()

    def test_decorator_explicit(self):
        @AuthException.recast(MyException1)
        def raise_my_exception1():
            raise MyException1(msg="test")

        with self.assertRaises(AuthException):
            raise_my_exception1()

    def test_passes_not_recast_exception(self):
        @AuthException.recast(MyException1)
        def raise_my_exception2():
            raise MyException2(msg="test")

        with self.assertRaises(MyException2):
            raise_my_exception2()

    def test_recast_child_exception(self):
        @MyAuthException.recast(MyException1)
        def raise_my_exception1():
            raise MyException1(msg="test raising MyException1 that should be recast to MyAuthException")

        with self.assertRaises(MyAuthException):
            raise_my_exception1()

    def test_multi_recast1(self):
        @MyAuthExceptionSub2.recast(MyException1)
        @MyAuthExceptionSub1.recast(MyException1Sub1)
        def raise_my_exception1sub1():
            raise MyException1Sub1(msg="Test")

        with self.assertRaises(MyAuthExceptionSub1):
            raise_my_exception1sub1()

    def test_multi_recast2(self):
        @MyAuthExceptionSub2.recast(MyException1)
        @MyAuthExceptionSub1.recast(MyException1Sub1)
        def raise_my_exception1():
            raise MyException1(msg="Test")

        with self.assertRaises(MyAuthExceptionSub2):
            raise_my_exception1()

    def test_multi_recast3(self):
        # @MyAuthException3.recast(Exception) Catching and recasting to something caught higher up the stack is problematic.
        @MyAuthException2.recast(MyException2)
        @MyAuthExceptionSub2.recast(MyException1)
        @MyAuthExceptionSub1.recast(MyException1Sub1)
        def raise_my_exception2():
            raise MyException2(msg="Test")

        with self.assertRaises(MyAuthException2):
            raise_my_exception2()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import unittest
from unittest import mock

from planet_auth import FileBackedOidcCredential, RefreshingOidcTokenRequestAuthenticator, FileBackedApiKey
from planet_auth.oidc.auth_clients.device_code_flow import (
    DeviceCodeClientConfig,
    DeviceCodeAuthClient,
    DeviceCodeAuthClientException,
)
from tests.test_planet_auth.unit.auth.util import StubOidcClientConfig, StubOidcAuthClient
from tests.test_planet_auth.util import tdata_resource_file_path, mock_sleep_skip

TEST_AUTH_SERVER = "https://blackhole.unittest.planet.com/fake_authserver"
TEST_CLIENT_ID = "fake_test_client_id"
TEST_TOKEN_SAVE_FILE_PATH = pathlib.Path("/test/token.json")
MOCK_TOKEN = {"token_type": "Bearer", "expires_in": 3600, "access_token": "__mock_access_token__", "scope": "planet"}
MOCK_DEVICE_AUTH_REQ_RESULT_WITH_COMPLETE_URI = {
    "device_code": "__mock_device_code__",
    "user_code": "__mock_user_code__",
    "verification_uri": "__mock_verification_uri__",
    "verification_uri_complete": "__mock_verification_uri_complete__",
    "expires_in": 100,
    "interval": 1,
}
MOCK_DEVICE_AUTH_REQ_RESULT_WITHOUT_COMPLETE_URI = {
    "device_code": "__mock_device_code__",
    "user_code": "__mock_user_code__",
    "verification_uri": "__mock_verification_uri__",
    "expires_in": 100,
    "interval": 1,
}

# Get some tokens that look more real than the mock token for some tests.
TEST_SIGNING_KEY_FILE = tdata_resource_file_path("keys/keypair1_priv_nopassword.test_pem")
TEST_SIGNING_PUBKEY_FILE = tdata_resource_file_path("keys/keypair1_pub_jwk.json")
STUB_AUTH_CLIENT_CONFIG = StubOidcClientConfig(
    auth_server=TEST_AUTH_SERVER,
    stub_authority_ttl=3600,
    stub_authority_access_token_audience="utest_audience",
    stub_authority_signing_key_file=TEST_SIGNING_KEY_FILE,
    stub_authority_pub_key_file=TEST_SIGNING_PUBKEY_FILE,
    scopes=["offline_access", "profile", "openid", "test_scope_1", "test_scope_2"],
)
STUB_AUTH_CLIENT = StubOidcAuthClient(STUB_AUTH_CLIENT_CONFIG)


def mocked_devauthapi_request_code_with_complete_uri(
    obj_self, client_id, requested_scopes, requested_audiences, auth_enricher, extra
):
    return MOCK_DEVICE_AUTH_REQ_RESULT_WITH_COMPLETE_URI


def mocked_devauthapi_request_code_without_complete_uri(
    obj_self, client_id, requested_scopes, requested_audiences, auth_enricher, extra
):
    return MOCK_DEVICE_AUTH_REQ_RESULT_WITHOUT_COMPLETE_URI


def mocked_tokenapi_from_device_code(obj_self, client_id, device_code, timeout, poll_interval, auth_enricher):
    return MOCK_TOKEN


class DeviceCodeNoClientAuthTest(unittest.TestCase):
    def setUp(self):
        self.under_test = DeviceCodeAuthClient(
            DeviceCodeClientConfig(
                auth_server=TEST_AUTH_SERVER,
                client_id=TEST_CLIENT_ID,
                token_endpoint=TEST_AUTH_SERVER + "/token",
                device_authorization_endpoint=TEST_AUTH_SERVER + "/device_auth",
            )
        )

    @mock.patch("time.sleep", mock_sleep_skip)
    @mock.patch(
        "planet_auth.oidc.api_clients.token_api_client.TokenApiClient.poll_for_token_from_device_code",
        mocked_tokenapi_from_device_code,
    )
    @mock.patch(
        "planet_auth.oidc.api_clients.device_authorization_api_client.DeviceAuthorizationApiClient.request_device_code",
        mocked_devauthapi_request_code_with_complete_uri,
    )
    def test_login_1(self):
        test_result = self.under_test.login(allow_tty_prompt=True)
        self.assertIsInstance(test_result, FileBackedOidcCredential)
        self.assertEqual(MOCK_TOKEN, test_result.data())

        # again with override scopes, but since the response is mocked
        # there is nothing different to check in the result data.
        # Same goes for audiences.
        test_result = self.under_test.login(
            allow_tty_prompt=True, requested_scopes=["override1"], requested_audiences=["req_aud1"]
        )
        self.assertIsInstance(test_result, FileBackedOidcCredential)
        # self.assertEqual(MOCK_TOKEN, test_result.data())

    @mock.patch("time.sleep", mock_sleep_skip)
    @mock.patch(
        "planet_auth.oidc.api_clients.token_api_client.TokenApiClient.poll_for_token_from_device_code",
        mocked_tokenapi_from_device_code,
    )
    @mock.patch(
        "planet_auth.oidc.api_clients.device_authorization_api_client.DeviceAuthorizationApiClient.request_device_code",
        mocked_devauthapi_request_code_without_complete_uri,
    )
    def test_login_2(self):
        # As above, but a slightly different response
        test_result = self.under_test.login(allow_tty_prompt=True)
        self.assertIsInstance(test_result, FileBackedOidcCredential)
        self.assertEqual(MOCK_TOKEN, test_result.data())

    def test_default_request_authenticator_type(self):
        test_result = self.under_test.default_request_authenticator(credential=TEST_TOKEN_SAVE_FILE_PATH)
        self.assertIsInstance(test_result, RefreshingOidcTokenRequestAuthenticator)

    def test_default_request_authenticator_given_literal_credential_object_instead_of_path(self):
        literal_credential = STUB_AUTH_CLIENT.login()
        test_result = self.under_test.default_request_authenticator(credential=literal_credential)
        test_result.pre_request_hook()
        self.assertEqual(literal_credential.access_token(), test_result._token_body)

    def test_default_request_authenticator_given_invalid_type(self):
        literal_credential = FileBackedApiKey()
        with self.assertRaises(TypeError):
            self.under_test.default_request_authenticator(credential=literal_credential)

    def test_login_fails_no_user_input_allowed(self):
        with self.assertRaises(DeviceCodeAuthClientException):
            self.under_test.login(allow_open_browser=False, allow_tty_prompt=False)


class DeviceCodeClientSecretClientAuthTest(unittest.TestCase):
    # No substance to test in this class
    pass


class DeviceCodeClientPubKeyClientAuthTest(unittest.TestCase):
    # No substance to test in this class
    pass



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

from cryptography.hazmat.primitives import serialization as crypto_serialization

from planet_auth.auth_client import AuthClientConfigException
from planet_auth.oidc.api_clients.api_client import EnricherPayloadType, EnricherReturnType
from planet_auth.oidc.api_clients.oidc_request_auth import prepare_private_key_assertion_auth_payload
from planet_auth.oidc.auth_client import OidcAuthClient, OidcAuthClientConfig


class OidcAuthClientWithPubKeyClientConfig(OidcAuthClientConfig, ABC):
    """
    Mix-in base class for OIDC auth clients that make use of a client public/private keypair
    to perform client authentication.
    """

    def __init__(
        self,
        client_privkey: str = None,
        client_privkey_file: str = None,
        client_privkey_password: str = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if client_privkey:
            self._data["client_privkey"] = client_privkey
        if client_privkey_file:
            self._data["client_privkey_file"] = client_privkey_file
        if client_privkey_password:
            self._data["client_privkey_password"] = client_privkey_password

        # Loaded JIT. Not in the serialized self._data representation.
        self._client_privkey_data = None

    def check_data(self, data):
        super().check_data(data)
        if not data.get("client_privkey") and not data.get("client_privkey_file"):
            raise AuthClientConfigException(
                message="One of client_privkey or client_privkey_file must be"
                " configured for {} client".format(self.__class__.__name__)
            )

    def client_privkey(self) -> str:
        return self.lazy_get("client_privkey")

    def client_privkey_data(self):
        # TODO: handle key refresh if the file has changed?
        if not self._client_privkey_data:
            self._load_private_key()
        return self._client_privkey_data

    def client_privkey_file(self) -> str:
        return self.lazy_get("client_privkey_file")

    def client_privkey_password(self) -> str:
        return self.lazy_get("client_privkey_password")

    # Recast is to catches bad passwords. Too broad?
    @AuthClientConfigException.recast(TypeError, ValueError)
    def _load_private_key(self):
        # TODO: also handle loading of JWK keys? Fork based on filename
        #       or detect?
        # import jwt
        # priv_key = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(key_dict))

        config_privkey_password = self.client_privkey_password()
        if config_privkey_password and type(config_privkey_password) is str:
            privkey_password = config_privkey_password.encode()
        else:
            privkey_password = config_privkey_password

        config_privkey_literal = self.client_privkey()
        if config_privkey_literal:
            if type(config_privkey_literal) is str:
                _privkey_literal = config_privkey_literal.encode()
            else:
                _privkey_literal = config_privkey_literal

            priv_key = crypto_serialization.load_pem_private_key(_privkey_literal, password=privkey_password)
            if not priv_key:
                raise AuthClientConfigException(
                    message="Unable to load private key literal from configuration for {} auth client.".format(
                        self.__class__.__name__
                    )
                )
        else:
            _config_privkey_file = self.client_privkey_file()
            if not _config_privkey_file:
                raise AuthClientConfigException(
                    message="Private key must be configured for {} auth client.".format(self.__class__.__name__)
                )
            with open(_config_privkey_file, "rb") as key_file:
                priv_key = crypto_serialization.load_pem_private_key(key_file.read(), password=privkey_password)
                if not priv_key:
                    raise AuthClientConfigException(
                        message='Unable to load private key from file "{}"'.format(_config_privkey_file)
                    )

        self._client_privkey_data = priv_key

    @classmethod
    def meta(cls):
        return {
            "config_hints": super().meta().get("config_hints")
            + [
                {
                    "config_key": "client_privkey",
                    "config_key_name": "Client private key literal",
                    "config_key_description": "A literal private key to store in the auth client configuration."
                    "  Clients require one of a private key literal or a private key file.",
                },
                {
                    "config_key": "client_privkey_file",
                    "config_key_name": "Client private key file",
                    "config_key_description": "Path to a file containing the client's private key."
                    "  Clients require one of a private key literal or a private key file.",
                },
                {
                    "config_key": "client_privkey_password",
                    "config_key_name": "Private key password",
                    "config_key_description": "Optional password protecting the private key",
                },
            ]
        }


class OidcAuthClientWithClientPubkey(OidcAuthClient, ABC):
    """
    Mix-in base class for OIDC auth clients that make use of a client public/private keypair
    to perform client authentication.
    """

    def __init__(self, client_config: OidcAuthClientWithPubKeyClientConfig):
        super().__init__(client_config)
        self._oidc_pubkey_client_config = client_config

    def _client_auth_enricher(self, raw_payload: EnricherPayloadType, audience: str) -> EnricherReturnType:
        auth_assertion_payload = prepare_private_key_assertion_auth_payload(
            audience=audience,
            client_id=self._oidc_pubkey_client_config.client_id(),
            private_key=self._oidc_pubkey_client_config.client_privkey_data(),
            ttl=300,
        )
        enriched_payload = {**raw_payload, **auth_assertion_payload}
        return enriched_payload, None



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.metadata

AUTH_CONFIG_FILE_PLAIN = "auth_client.json"
AUTH_CONFIG_FILE_SOPS = "auth_client.sops.json"
TOKEN_FILE_PLAIN = "token.json"
TOKEN_FILE_SOPS = "token.sops.json"
USER_CONFIG_FILE = ".planet.json"
PROFILE_DIR = ".planet"
X_PLANET_APP = f"planet-auth-library-{importlib.metadata.version('planet-auth')}"
X_PLANET_APP_HEADER = "X-Planet-App"



import requests

from planet_auth import Auth
import planet_auth_config  # type: ignore


def main():
    _legacy_client_config = {
        **planet_auth_config.Staging.LEGACY_AUTH_AUTHORITY,
        "client_type": "planet_legacy",
    }

    auth_ctx = Auth.initialize_from_config_dict(client_config=_legacy_client_config, token_file="/saved_token.json")
    # The presumption in this example is that the login has already been performed and the token saved.
    # See the login example for that process.

    result = requests.get(url="https://api.planet.com/basemaps/v1/mosaics", auth=auth_ctx.request_authenticator())
    print(result.status_code)
    print(result.json())


if __name__ == "__main__":
    main()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from typing import Dict, List, Optional

import planet_auth.logging.auth_logger
from planet_auth.oidc.api_clients.api_client import (
    OidcApiClient,
    OidcApiClientException,
    EnricherFuncType,
    _RequestParamsType,
)

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class TokenApiException(OidcApiClientException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class TokenApiTimeoutException(TokenApiException):
    pass


class TokenApiClient(OidcApiClient):
    """
    Low level API client for the OAuth Token endpoint. See
    RFC 6749 - The OAuth 2.0 Authorization Framework for protocol details.
    """

    def __init__(self, token_uri: str):
        """
        Create a new Token API Client.
        """
        super().__init__(endpoint_uri=token_uri)

    @staticmethod
    def _check_valid_token_response(json_response: Dict) -> None:
        if not json_response.get("expires_in"):
            # https://datatracker.ietf.org/doc/html/rfc6749#section-5.1
            # Note: while OAuth requires the access_token field, and OIDC
            # built on top of oauth, it's addition of an ID token option
            # means that there might not be an access toke in a successful
            # response.
            #
            # 'expires_in' is only recommended in JWTs, not required, by the OAuth
            # spec. (But, I think it's required in the "token" response,
            # which is distinct from the JWT payload.)
            raise TokenApiException(message="Invalid token received. Missing expires_in field.")
            # auth_logger.warning(msg='Token response was missing expires_in field.')

    def _checked_call(
        self, token_params: _RequestParamsType, auth_enricher: Optional[EnricherFuncType] = None
    ) -> Dict:
        request_auth = None
        if auth_enricher:
            token_params, request_auth = auth_enricher(token_params, self._endpoint_uri)

        json_response = self._checked_post_json_response(token_params, request_auth)
        self._check_valid_token_response(json_response)
        return json_response

    def _polling_checked_call(
        self,
        token_params: _RequestParamsType,
        timeout: float,
        poll_interval: float,
        auth_enricher: Optional[EnricherFuncType] = None,
    ) -> Dict:
        start_time = time.time()
        while True:
            try:
                json_response = self._checked_call(token_params=token_params, auth_enricher=auth_enricher)
                return json_response
            except OidcApiClientException as oe:
                if oe.error_code == "authorization_pending":
                    pass
                elif oe.error_code == "slow_down":
                    poll_interval += 5  # See RFC 8628
                else:
                    raise oe
            now_time = time.time()
            if (now_time - start_time) < timeout:
                time.sleep(poll_interval)
            else:
                # We expect a expired_token error code, but the caller could
                # indicate they don't want to wait as long as the server is willing to.
                raise TokenApiTimeoutException(message="Timeout exceeded")

    def get_token_from_refresh(
        self,
        client_id: str,
        refresh_token: str,
        requested_scopes: List[str] = None,
        auth_enricher: Optional[EnricherFuncType] = None,
        extra: Dict = None,
    ) -> Dict:
        """
        Obtain tokens using a refresh token.

        Parameters:
            client_id: The ID of the OAuth client making the request.
            refresh_token: The refresh token to use.
            requested_scopes: A list of scopes to request when obtaining the
                refreshed tokens.  This list may only be a subset of the scopes
                that were initially granted to the client.  A scope increase may
                not be requested during refresh. A scope increase requires a new
                authorization (login).
            extra: Dict of extra parameters to pass to the token endpoint.

        Returns:
            Returns the token endpoint response payload upon success as a json
                object. If an error condition is detected, either from the HTTP
                layer or from a payload that indicates a failure, an exception
                will be raised.
        """
        if extra is None:
            extra = {}
        # "None" is pythonic, and does not mean anything to OAuth APIs.
        _extra = {k: v for k, v in extra.items() if v is not None}
        data = {
            **_extra,  # we do this first, because we want it to lose the race with explicit parameters. It is not intended for clobbering.
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            # Client id, secret, or assertion come from enricher.
            # 'client_id': client_id,
            # 'client_secret': client_secret
            # 'client_assertion_type':
            #        'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            #  client_assertion': signed_jwt
        }
        if requested_scopes:
            # You can request down-scope on refresh, but the refresh token
            # remains potent.
            data["scope"] = " ".join(requested_scopes)

        return self._checked_call(data, auth_enricher)

    def get_token_from_client_credentials(
        self,
        client_id: str,
        requested_scopes: List[str] = None,
        requested_audiences: List[str] = None,
        auth_enricher: Optional[EnricherFuncType] = None,
        extra: Dict = None,
    ) -> Dict:
        """
        Obtain tokens using client credentials and the client credentials grant
        OAuth flow.

        Parameters:
            client_id: The ID of the OAuth client making the request.
                (This parameter is not used. This comes from the auth enricher.)
            requested_audiences: A list of strings specifying the audiences
                to request.
            requested_scopes: A list of scopes to request when obtaining the
                refreshed tokens.  This list may only be a subset of the scopes
                that were initially granted to the client.  A scope increase may
                not be requested during refresh. A scope increase requires a new
                authorization (login).
            auth_enricher: Function to layer in the application of client credentials.
            extra: Dict of extra parameters to pass to the token endpoint.

        Returns:
            Returns the token endpoint response payload upon success as a json
                object. If an error condition is detected, either from the HTTP
                layer or from a payload that indicates a failure, an exception
                will be raised.
        """
        if extra is None:
            extra = {}
        # "None" is pythonic, and does not mean anything to OAuth APIs.
        _extra = {k: v for k, v in extra.items() if v is not None}
        data = {
            **_extra,  # we do this first, because we want it to lose the race with explicit parameters. It is not intended for clobbering.
            "grant_type": "client_credentials",
            # Client id, secret, or assertion come from enricher.
            # 'client_id': client_id,
            # 'client_secret': client_secret
            # 'client_assertion_type':
            #        'urn:ietf:params:oauth:client-assertion-type:jwt-bearer',
            #  client_assertion': signed_jwt
        }
        if requested_scopes:
            data["scope"] = " ".join(requested_scopes)
        if requested_audiences:
            data["audience"] = " ".join(requested_audiences)

        return self._checked_call(data, auth_enricher)

    def get_token_from_code(
        self,
        redirect_uri: str,
        client_id: str,
        code: str,
        code_verifier: str,
        auth_enricher: Optional[EnricherFuncType] = None,
        # extra=None, # This should be in the auth request, not the code redemption.
    ) -> Dict:
        """
        Obtain tokens using an authorization code.

        Returns:
            Returns the token endpoint response payload upon success as a json
                object. If an error condition is detected, either from the HTTP
                layer or from a payload that indicates a failure, an exception
                will be raised.
        """
        # if extra is None:
        #     extra = {}
        # "None" is pythonic, and does not mean anything to OAuth APIs.
        # _extra = {k: v for k, v in extra.items() if v is not None}
        data = {
            # **_extra,  # we do this first, because we want it to lose the race with explicit parameters. It is not intended for clobbering.
            "grant_type": "authorization_code",
            "redirect_uri": redirect_uri,
            "code": code,
            "code_verifier": code_verifier,
            # 'client_id': client_id,  # The job of the enricher.
        }

        return self._checked_call(data, auth_enricher)

    def poll_for_token_from_device_code(
        self,
        client_id: str,
        device_code: str,
        timeout: int,
        poll_interval: int = 5,  # Default poll interval specified in RFC 8628
        auth_enricher: Optional[EnricherFuncType] = None,
        # extra=None,  # This should be in the auth request, not the code redemption.
    ) -> Dict:
        """
        Poll for the completion of a device code login.

        Returns:
            Returns the token endpoint response payload upon success as a json
                object. If an error condition is detected, either from the HTTP
                layer or from a payload that indicates a failure, an exception
                will be raised.
        """
        # if extra is None:
        #    extra = {}
        # "None" is pythonic, and does not mean anything to OAuth APIs.
        # _extra = {k: v for k, v in extra.items() if v is not None}
        data = {
            # **_extra,  # we do this first, because we want it to lose the race with explicit parameters. It is not intended for clobbering.
            "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
            "device_code": device_code,
            # 'client_id': client_id,  # The job of the enricher.
        }
        return self._polling_checked_call(
            data, timeout=timeout, poll_interval=poll_interval, auth_enricher=auth_enricher
        )

    def get_token_from_password(
        self,
        client_id: str,
        username: str,
        password: str,
        requested_scopes: List[str] = None,
        requested_audiences: List[str] = None,
        auth_enricher: Optional[EnricherFuncType] = None,
        extra: Dict = None,
    ) -> Dict:
        """
        Obtain tokens using a username and password and the resource owner
        grant OAuth flow.

        Parameters:
            client_id: The ID of the OAuth client making the request.
                (This parameter is not used. This comes from the auth enricher.)
            username: The username used for authentication.
            password: The password used for authentication.
            requested_scopes: A list of scopes to request when obtaining the
                refreshed tokens.  This list may only be a subset of the scopes
                that were initially granted to the client.  A scope increase may
                not be requested during refresh. A scope increase requires a new
                authorization (login).
            requested_audiences: A list of strings specifying the audiences
                to request.
            auth_enricher: Function to layer un the application of client credentials.
            extra: Dict of extra parameters to pass to the token endpoint.

        Returns:
            Returns the token endpoint response payload upon success as a json
                object. If an error condition is detected, either from the HTTP
                layer or from a payload that indicates a failure, an exception
                will be raised.
        """
        if extra is None:
            extra = {}
        # "None" is pythonic, and does not mean anything to OAuth APIs.
        _extra = {k: v for k, v in extra.items() if v is not None}
        data = {
            **_extra,  # we do this first, because we want it to lose the race with explicit parameters. It is not intended for clobbering.
            "grant_type": "password",
            "username": username,
            "password": password,
            # "client_id": client_id, The job of the auth_enricher.
        }
        if requested_scopes:
            data["scope"] = " ".join(requested_scopes)
        if requested_audiences:
            data["audience"] = " ".join(requested_audiences)

        return self._checked_call(data, auth_enricher)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import shutil
import tempfile
import pathlib

from planet_auth.planet_legacy.legacy_api_key import FileBackedPlanetLegacyApiKey
from planet_auth.planet_legacy.request_authenticator import PlanetLegacyRequestAuthenticator
from planet_auth.storage_utils import FileBackedJsonObjectException
from tests.test_planet_auth.util import tdata_resource_file_path


class PlanetLegacyRequestAuthenticatorTest(unittest.TestCase):
    def setUp(self):
        # Copy because some actions may modify the test data files.
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.tmp_dir_path = pathlib.Path(self.tmp_dir.name)

        self.invalid_cred_file = self.tmp_dir_path.joinpath("invalid_test_credential.json")
        shutil.copy(
            tdata_resource_file_path("keys/invalid_test_credential.json"),
            self.invalid_cred_file,
        )

        self.valid_cred_file = self.tmp_dir_path.joinpath("planet_legacy_test_credential.json")
        shutil.copy(
            tdata_resource_file_path("keys/planet_legacy_test_credential.json"),
            self.valid_cred_file,
        )

    def test_pre_request_hook_loads_from_file_happy_path(self):
        under_test = PlanetLegacyRequestAuthenticator(
            planet_legacy_credential=FileBackedPlanetLegacyApiKey(api_key_file=self.valid_cred_file)
        )
        under_test.pre_request_hook()
        self.assertEqual("test_legacy_api_key", under_test._api_key_file.legacy_api_key())

    def test_pre_request_hook_loads_from_file_invalid_throws(self):
        under_test = PlanetLegacyRequestAuthenticator(
            planet_legacy_credential=FileBackedPlanetLegacyApiKey(api_key_file=self.invalid_cred_file)
        )
        with self.assertRaises(FileBackedJsonObjectException):
            under_test.pre_request_hook()

    def test_update_credential_data(self):
        under_test = PlanetLegacyRequestAuthenticator(
            planet_legacy_credential=FileBackedPlanetLegacyApiKey(api_key_file=self.valid_cred_file)
        )
        under_test.pre_request_hook()  # Triggers a JIT load from the file

        orig_data = under_test._credential.data()
        new_data = orig_data.copy()
        new_data["key"] = "utest_new_api_key"

        under_test.update_credential_data(new_credential_data=new_data)
        # Side effect : clobbers the file contents.

        under_test.pre_request_hook()  # Simulate another login event
        self.assertEqual("utest_new_api_key", under_test._token_body)  # Base class view
        self.assertEqual("utest_new_api_key", under_test._api_key_file.legacy_api_key())  # child class view

        # self.assertEqual("TODO?", under_test._token_prefix)

    def test_update_credential(self):
        under_test = PlanetLegacyRequestAuthenticator(
            planet_legacy_credential=FileBackedPlanetLegacyApiKey(api_key_file=self.valid_cred_file)
        )
        under_test.pre_request_hook()  # Triggers a JIT load from the file of the initial credential.

        new_credential = FileBackedPlanetLegacyApiKey(api_key="utest_new_api_key_2")
        under_test.update_credential(new_credential=new_credential)

        under_test.pre_request_hook()  # Simulate another login event
        self.assertEqual("utest_new_api_key_2", under_test._token_body)  # Base class view
        self.assertEqual("utest_new_api_key_2", under_test._api_key_file.legacy_api_key())  # child class view

        # self.assertEqual("TODO?", under_test._token_prefix)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib.resources
import os
import pathlib
import tempfile

from planet_auth_utils.profile import Profile


def tdata_resource_file_path(resource_file: str):
    file_path = importlib.resources.files("tests.test_planet_auth_utils").joinpath("data/" + resource_file)
    return file_path


class TestWithHomeDirProfiles:
    def setUp_testHomeDir(self):
        # pylint: disable=W0201
        self.test_home_dir = tempfile.TemporaryDirectory()
        self.test_home_dir_path = pathlib.Path(self.test_home_dir.name)
        self.old_home = os.environ.get("HOME")

        self.planet_dir_path = self.test_home_dir_path.joinpath(".planet")
        self.planet_dir_path.mkdir()

        os.environ["HOME"] = self.test_home_dir.name

    def tearDown_testHomeDir(self):
        if self.old_home:
            os.environ["HOME"] = self.old_home
        else:
            os.environ.pop("HOME")

    def mkProfileDir(self, profile_name) -> pathlib.Path:
        # Mimic the behavior of the default storage provider
        new_profile_dir_path = pathlib.Path.home() / Profile.get_profile_dir_path(profile_name)
        # new_profile_dir_path = Profile.get_profile_dir_path(profile_name)
        new_profile_dir_path.mkdir()
        return new_profile_dir_path



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import List, Union

import planet_auth.logging.auth_logger
from planet_auth.auth_exception import AuthException
from planet_auth.auth_client import AuthClientConfig
from planet_auth.constants import (
    AUTH_CONFIG_FILE_SOPS,
    AUTH_CONFIG_FILE_PLAIN,
    PROFILE_DIR,
)
from planet_auth.storage_utils import ObjectStorageProvider


auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


class ProfileException(AuthException):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class Profile:
    """
    Tools for managing configuration within a profile directory on disk.
    """

    @staticmethod
    def profile_root() -> pathlib.Path:
        """
        Root storage directory used for profile data.
        """
        # return pathlib.Path.home().joinpath(PROFILE_DIR)
        # We used to assume file storage.  We now support pluggable storage implementation.
        # We let the storage provider determine the real path (if there is one) and in this
        # Profile class we now (mostly) deal in the abstract idea of the path as an object
        # identifier that may or may not be file system based, depending on the storage
        # provider in use.
        return pathlib.Path(PROFILE_DIR)

    @staticmethod
    def get_profile_dir_path(profile: str) -> pathlib.Path:
        return Profile.profile_root().joinpath(profile.lower())

    @staticmethod
    def get_profile_file_path(
        filename: str, profile: str, override_path: Union[str, pathlib.PurePath, None] = None
    ) -> pathlib.Path:
        """
        Given a profile name and a file name, construct a file path for
        the file under the profile directory.  If an override is given,
        it will always be chosen.
        """
        if override_path:
            return pathlib.Path(override_path)

        if not profile or profile == "":
            raise ProfileException(message="profile must be set")

        return Profile.get_profile_dir_path(profile).joinpath(filename)

    @staticmethod
    def get_profile_file_path_with_priority(
        filenames: List[str],
        profile: str,
        override_path: Union[str, pathlib.PurePath, None] = None,
        # storage_provider: Optional[ObjectStorageProvider] = None,  # not yet supported here.
    ) -> pathlib.Path:
        """
        Given a list of candidate filenames, choose the first that that
        exists. If none exist, the last one will be used regardless of whether the
        file exists or not.  If the application needs the first rather than the
        last entry to be the fallback value, it should repeat that value and the
        end of the list.
        """
        if override_path:
            return pathlib.Path(override_path)

        last_candidate_path = None
        for candidate_filename in filenames:
            candidate_path = Profile.get_profile_file_path(
                filename=candidate_filename, profile=profile, override_path=None
            )
            last_candidate_path = candidate_path
            # TODO: custom storage providers not yet supported in this path
            _storage_provider = ObjectStorageProvider._default_storage_provider()
            if _storage_provider.obj_exists(candidate_path):
                return candidate_path

        return last_candidate_path  # type: ignore

    @staticmethod
    def load_auth_client_config(
        profile: str,
        # storage_provider: Optional[ObjectStorageProvider] = None,  # not yet supported here.
    ) -> AuthClientConfig:
        auth_config_path = Profile.get_profile_file_path_with_priority(
            filenames=[AUTH_CONFIG_FILE_SOPS, AUTH_CONFIG_FILE_PLAIN],
            profile=profile,
        )
        # TODO: custom storage providers not yet supported in this path
        _storage_provider = ObjectStorageProvider._default_storage_provider()
        if _storage_provider.obj_exists(auth_config_path):
            auth_logger.debug(msg='Using auth client configuration from "{}"'.format(str(auth_config_path)))
            client_config = AuthClientConfig.from_file(file_path=auth_config_path, storage_provider=_storage_provider)
        else:
            raise FileNotFoundError('Auth configuration file "{}" not found.'.format(str(auth_config_path)))

        return client_config

    @staticmethod
    def list_on_disk_profiles() -> List[str]:
        # Any directory in ~/.planet is only a potential profile. We only
        # consider it an actual profile if a client config file can be found.
        # Whether or not a profile is valid is not considered.

        # TODO: This assumes the default storage provider default behavior, since
        #   The placement under is now set by the storage provider, and not the Profile class.
        profile_abs_dir = pathlib.Path.home() / Profile.profile_root()
        if not profile_abs_dir.is_dir():
            return []
        candidate_profile_names = [x.name for x in profile_abs_dir.iterdir() if x.is_dir()]
        profile_names = []
        for candidate_profile_name in candidate_profile_names:
            config_file = Profile.get_profile_file_path_with_priority(
                profile=candidate_profile_name,
                filenames=[AUTH_CONFIG_FILE_SOPS, AUTH_CONFIG_FILE_PLAIN],
                override_path=None,
            )
            # TODO: custom storage providers not yet supported in this path
            _storage_provider = ObjectStorageProvider._default_storage_provider()
            if _storage_provider.obj_exists(config_file):
                profile_names.append(candidate_profile_name)

        profile_names.sort()
        return profile_names
